/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrenderdrawable
 * @short_description: Base abstract class for drawable objects.
 *
 * #PgmRenderDrawable is a base abstract class for drawable renderer classes.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrenderdrawable.h"
#include <gst/gst.h>

enum {
  ARG_0,
  ARG_CAPS
};

static void pgm_render_drawable_set_property
  (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void pgm_render_drawable_get_property
  (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void pgm_render_drawable_class_init (PgmRenderDrawableClass *klass);
static void pgm_render_drawable_init (PgmRenderDrawable *object);
static void pgm_render_drawable_finalize (GObject *object);

static GObjectClass *parent_class = NULL;

G_DEFINE_ABSTRACT_TYPE (PgmRenderDrawable, pgm_render_drawable, G_TYPE_OBJECT);

static void
pgm_render_drawable_set_property (GObject *object, guint prop_id,
                                  const GValue *value, GParamSpec *pspec)
{
  PgmRenderDrawable *drawable;

  g_return_if_fail (PGM_RENDER_IS_DRAWABLE (object));
  drawable = PGM_RENDER_DRAWABLE (object);

  switch (prop_id)
    {
    case ARG_CAPS:
      PGM_RENDER_DRAWABLE_GET_CLASS (drawable)->caps =
        /* GST_CAPS  */(g_value_get_pointer (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
pgm_render_drawable_get_property (GObject *object, guint prop_id, GValue *value,
                                  GParamSpec *pspec)
{
  PgmRenderDrawable *drawable;

  g_return_if_fail (PGM_RENDER_IS_DRAWABLE (object));
  drawable = PGM_RENDER_DRAWABLE (object);

  switch (prop_id)
    {
    case ARG_CAPS:
      g_value_set_pointer (value,
                           PGM_RENDER_DRAWABLE_GET_CLASS (drawable)->caps);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
pgm_render_drawable_class_init (PgmRenderDrawableClass *klass)
{
  GObjectClass *gobject_class;
  gobject_class = (GObjectClass*) klass;

  parent_class = g_type_class_peek_parent (klass);
  gobject_class->finalize = pgm_render_drawable_finalize;
  gobject_class->set_property = pgm_render_drawable_set_property;
  gobject_class->get_property = pgm_render_drawable_get_property;

  g_object_class_install_property
    (gobject_class, ARG_CAPS, g_param_spec_pointer
     ("caps", "Caps", "The capacities of the drawable",
      G_PARAM_READWRITE));

  klass->caps = NULL;
}

static void
pgm_render_drawable_init (PgmRenderDrawable *object)
{
  object->lock = g_new0 (GStaticMutex, 1);
  g_static_mutex_init (object->lock);
  object->window = NULL;
  object->order = PGM_RENDER_DRAWABLE_UNBINDED;
  object->show = TRUE;
}

static void
pgm_render_drawable_finalize (GObject *object)
{
  PgmRenderDrawable *drawable = PGM_RENDER_DRAWABLE (object);

  g_static_mutex_free (drawable->lock);
  g_free (drawable->lock);
  drawable->lock = NULL;
  drawable->window = NULL;

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * pgm_render_drawable_new:
 *
 * Creates a new #PgmRenderDrawable instance.
 *
 * Returns: A new #PgmRenderDrawable.
 */
PgmRenderDrawable *
pgm_render_drawable_new (void)
{
  return PGM_RENDER_DRAWABLE (g_object_new (PGM_RENDER_TYPE_DRAWABLE, NULL));
}

/**
 * pgm_render_drawable_get_caps:
 * @drawable: A #PgmRenderDrawable object.
 * @caps: The caps string to fill.
 *
 * Retrieves the capabilities of the #drawable class.
 *
 * MT safe.
 *
 * Returns: A new #PgmRenderDrawable.
 */
PgmRenderError
pgm_render_drawable_get_caps (PgmRenderDrawable *drawable, GstCaps **caps)
{
  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  *caps = PGM_RENDER_DRAWABLE_GET_CLASS (drawable)->caps;

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_drawable_draw:
 * @drawable: A #PgmRenderDrawable object.
 *
 * Draw the object .
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_drawable_draw (PgmRenderDrawable *drawable)
{
  PgmRenderDrawableClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_DRAWABLE_GET_CLASS (drawable);

  if (klass->draw)
    klass->draw (drawable);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_drawable_hide:
 * @drawable: A #PgmRenderDrawable object.
 *
 * Hide the object.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_drawable_hide (PgmRenderDrawable *drawable)
{
  PgmRenderDrawableClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  drawable->show = FALSE;

  klass = PGM_RENDER_DRAWABLE_GET_CLASS (drawable);

  if (klass->hide)
    klass->hide (drawable);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_drawable_show:
 * @drawable: A #PgmRenderDrawable object.
 *
 * Show the object.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_drawable_show (PgmRenderDrawable *drawable)
{
  PgmRenderDrawableClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  drawable->show = TRUE;

  klass = PGM_RENDER_DRAWABLE_GET_CLASS (drawable);

  if (klass->show)
    klass->show (drawable);

  return PGM_RENDER_ERROR_X;
}
