/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrenderlayout-iface
 * @short_description: An interface to layout objects.
 *
 * #PgmRenderLayoutIFace is an interface to add different layout
 * capabilities to an object.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrenderlayout-iface.h"

static void pgm_render_layout_iface_base_init (gpointer g_class);

static void
pgm_render_layout_iface_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (! initialized)
    {
      initialized = TRUE;
    }
}

GType
pgm_render_layout_iface_get_type (void)
{
  static GType type = 0;

  if (! type)
    {
      static const GTypeInfo info = {
	sizeof (PgmRenderLayoutIFaceClass),
	pgm_render_layout_iface_base_init, /* base init */
	NULL, /* base finalize */
	NULL, /* class_init */
	NULL, /* class finalize */
	NULL, /* class data */
	0,
	0,    /* n_preallocs */
	NULL, /* instance init */
      };
      type = g_type_register_static (G_TYPE_INTERFACE, "PgmRenderLayoutIFace",
				     &info, 0);
      g_type_interface_add_prerequisite (type, G_TYPE_OBJECT);
    }

  return type;
}

/**
 * pgm_render_layout_set_size:
 * @layout: A #PgmRenderLayoutIFace object.
 * @width: The width of the object.
 * @height: The height of the object.
 *
 * Set the size of the object to (#width, #height).
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_set_size (PgmRenderLayoutIFace *layout, gint width,
			    gint height)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->set_size (layout, width, height);
}

/**
 * pgm_render_layout_get_size:
 * @layout: A #PgmRenderLayoutIFace object.
 * @width: The width of the object to fill.
 * @height: The height of the object to fill.
 *
 * Retrieves the (#width, #height) size of the object.
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_get_size (PgmRenderLayoutIFace *layout, gint *width,
                            gint *height)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout), PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->get_size (layout, width, height);
}

/**
 * pgm_render_layout_set_position:
 * @layout: A #PgmRenderLayoutIFace object.
 * @x: The x position of the object.
 * @y: The y position of the object.
 *
 * Set the position of the object to (#x, #y).
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_set_position (PgmRenderLayoutIFace *layout, gint x, gint y)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->set_position (layout, x, y);
}

/**
 * pgm_render_layout_get_position:
 * @layout: A #PgmRenderLayoutIFace object.
 * @x: The x position of the object to fill.
 * @y: The y position of the object to fill.
 *
 * Retrieves the (#x, #y) position of the object.
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_get_position (PgmRenderLayoutIFace *layout, gint *x, gint *y)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->get_position (layout, x, y);
}

/**
 * pgm_render_layout_set_alignment:
 * @layout: A #PgmRenderLayoutIFace object.
 * @x: The x alignment of the object between 0.0 and 1.0.
 * @y: The y alignment of the object between 0.0 and 1.0.
 *
 * Set the alignment of the object to (#x, #y).
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_set_alignment (PgmRenderLayoutIFace *layout, gfloat x,
				 gfloat y)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->set_alignment (layout, x, y);
}

/**
 * pgm_render_layout_get_alignment:
 * @layout: A #PgmRenderLayoutIFace object.
 * @x: The x alignment of the object to fill. 0.5 by default.
 * @y: The y alignment of the object to fill. 0.5 by default.
 *
 * Retrieves the (#x, #y) position of the object.
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_get_alignment (PgmRenderLayoutIFace *layout, gfloat *x,
				 gfloat *y)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->get_alignment (layout, x, y);
}

/**
 * pgm_render_layout_set_aspect_style:
 * @layout: A #PgmRenderLayoutIFace object.
 * @style: The aspect-ratio style of the object.
 *
 * Set the aspect-ratio #style of the object.
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_set_aspect_style (PgmRenderLayoutIFace *layout,
				    PgmRenderLayoutAspectStyle style)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->set_aspect_style (layout, style);
}

/**
 * pgm_render_layout_get_aspect_style:
 * @layout: A #PgmRenderLayoutIFace object.
 * @style: The aspect-ratio style of the object to fill.
 *
 * Retrieves the aspect-ratio #style of the object.
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_get_aspect_style (PgmRenderLayoutIFace *layout,
				    PgmRenderLayoutAspectStyle *style)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->get_aspect_style (layout, style);
}

/**
 * pgm_render_layout_set_aspect_ratio:
 * @layout: A #PgmRenderLayoutIFace object.
 * @ratio: The custom ratio.
 *
 * Set a custom aspect-ratio (width/height) to the object.
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_set_aspect_ratio (PgmRenderLayoutIFace *layout, gfloat ratio)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->set_aspect_ratio (layout, ratio);
}

/**
 * pgm_render_layout_get_aspect_ratio:
 * @layout: A #PgmRenderLayoutIFace object.
 * @ratio: The custom ratio to fill.
 *
 * Retrieves the custom aspect-ratio (width/height) of the object.
 *
 * Return value: 
 */
PgmRenderError
pgm_render_layout_get_aspect_ratio (PgmRenderLayoutIFace *layout,
				    gfloat *ratio)
{
  g_return_val_if_fail (PGM_RENDER_IS_LAYOUT_IFACE (layout),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_LAYOUT_IFACE_GET_CLASS
    (layout)->get_aspect_ratio (layout, ratio);
}
