/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrendermedia-iface
 * @short_description: An interface for drawing color and images.
 *
 * #PgmRenderMediaIFace is an interface to add color and images drawing
 * capabilities to an object.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrendermedia-iface.h"
#include <gst/gst.h>

static void pgm_render_media_iface_base_init (gpointer g_class);

static void
pgm_render_media_iface_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (! initialized)
    {
      initialized = TRUE;
    }
}

GType
pgm_render_media_iface_get_type (void)
{
  static GType type = 0;

  if (! type)
    {
      static const GTypeInfo info = {
	sizeof (PgmRenderMediaIFaceClass),
	pgm_render_media_iface_base_init, /* base init */
	NULL, /* base finalize */
	NULL, /* class_init */
	NULL, /* class finalize */
	NULL, /* class data */
	0,
	0,    /* n_preallocs */
	NULL, /* instance init */
      };
      type = g_type_register_static (G_TYPE_INTERFACE, "PgmRenderMediaIFace",
				     &info, 0);
      g_type_interface_add_prerequisite (type, G_TYPE_OBJECT);
    }

  return type;
}

/**
 * pgm_render_media_set_color:
 * @media: A #PgmRenderMediaIFace object.
 * @r: The R color layer.
 * @g: The G color layer.
 * @b: The B color layer.
 * @a: The A color layer.
 *
 * Sets the color of #media to (#r, #g, #b, #a).
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_set_color (PgmRenderMediaIFace *media, guchar r, guchar g,
			    guchar b, guchar a)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS
    (media)->set_color (media, r, g, b, a);
}

/**
 * pgm_render_media_get_color:
 * @media: A #PgmRenderMediaIFace object.
 * @r: The R color layer to fill.
 * @g: The G color layer to fill.
 * @b: The B color layer to fill.
 * @a: The A color layer to fill.
 *
 * Retrieves the color of #media.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_get_color (PgmRenderMediaIFace *media, guchar *r, guchar *g,
			    guchar *b, guchar *a)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS
    (media)->get_color (media, r, g, b, a);
}

/**
 * pgm_render_media_set_image:
 * @media: A #PgmRenderMediaIFace object.
 * @format: The format of pixels of the image.
 * @width: The width of the image.
 * @height: The height of the image.
 * @stride: The number of bytes between the start of rows in the
 *          image. Having this specified separate from width handles
 *          the padding at the end of rows.
 * @data: A pointer to the data supplied by the application.
 *
 * Set the image of the #media. The image is drawn over the color of
 * the #media.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_set_image (PgmRenderMediaIFace *media,
			    PgmRenderMediaPixelFormat format,
			    gint width, gint height, gint stride,
			    gconstpointer data)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS
    (media)->set_image (media, format, width, height, stride, data);
}

/**
 * pgm_render_media_set_gst_image:
 * @media: A #PgmRenderMediaIFace object.
 * @format: The format of pixels of the image.
 * @width: The width of the image.
 * @height: The height of the image.
 * @stride: The number of bytes between the start of rows in the
 *          image. Having this specified separate from width handles
 *          the padding at the end of rows.
 * @data: A pointer to a #GstBuffer.
 *
 * Set the GStreamer buffer of the #media. The buffer is drawn over
 * the color of the #media. The #GstBuffer is unref'd by Render.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_set_gst_image (PgmRenderMediaIFace *media,
				PgmRenderMediaPixelFormat format, gint width,
				gint height, gint stride, GstBuffer *buffer)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS
    (media)->set_gst_image (media, format, width, height, stride, buffer);
}

/**
 * pgm_render_media_set_image_from_drawable:
 * @media: A #PgmRenderMediaIFace object.
 * @drawable: The surface to copy the data from.
 *
 * Set the image of the #media from another #drawable object
 * implementing the MediaIface interface. The goal is to avoid video
 * memory duplication.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_set_image_from_drawable (PgmRenderMediaIFace *media,
					  PgmRenderDrawable *drawable)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS
    (media)->set_image_from_drawable (media, drawable);
}

/**
 * pgm_render_media_del_image:
 * @media: A #PgmRenderMediaIFace object.
 *
 * Delete the image of the @media.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_del_image (PgmRenderMediaIFace *media)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS (media)->del_image (media);
}

/**
 * pgm_render_media_set_interp_type:
 * @media: A #PgmRenderMediaIFace object.
 * @type: The interpolation type.
 *
 * Set the filter #type for the scaling interpolation of the #media's
 * image.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_set_interp_type (PgmRenderMediaIFace *media,
				  PgmRenderMediaInterpType type)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS
    (media)->set_interp_type (media, type);
}

/**
 * pgm_render_media_set_interp_type:
 * @media: A #PgmRenderMediaIFace object.
 * @type: The interpolation type to fill.
 *
 * Retrieves the filter type for the scaling interpolation of the #media's
 * image.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_media_get_interp_type (PgmRenderMediaIFace *media,
				  PgmRenderMediaInterpType *type)
{
  g_return_val_if_fail (PGM_RENDER_IS_MEDIA_IFACE (media), PGM_RENDER_ERROR_X);

  return PGM_RENDER_MEDIA_IFACE_GET_CLASS
    (media)->get_interp_type (media, type);
}
