/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrendershadow-iface
 * @short_description: An interface for drawing drop shadows.
 *
 * #PgmRenderShadowIFace is an interface to add different shadow drawing
 * capabilities to an object.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrendershadow-iface.h"

static void pgm_render_shadow_iface_base_init (gpointer g_class);

static void
pgm_render_shadow_iface_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (! initialized)
    {
      initialized = TRUE;
    }
}

GType
pgm_render_shadow_iface_get_type (void)
{
  static GType type = 0;
  
  if (! type)
    {
      static const GTypeInfo info = {
	sizeof (PgmRenderShadowIFaceClass),
	pgm_render_shadow_iface_base_init, /* base init */
	NULL, /* base finalize */
	NULL, /* class_init */
	NULL, /* class finalize */
	NULL, /* class data */
	0,
	0,    /* n_preallocs */
	NULL, /* instance init */
      };
      type = g_type_register_static (G_TYPE_INTERFACE, "PgmRenderShadowIFace",
				     &info, 0);
      g_type_interface_add_prerequisite (type, G_TYPE_OBJECT);
    }

  return type;
}

/**
 * pgm_render_shadow_set_radius:
 * @shadow: A #PgmRenderShadowIFace object.
 * @radius: The radius of the shadow. It is the width added to each
 *          side of the surface.
 *
 * A shadow could be drawn under an object. Sets the #radius of
 * #shadow.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_set_radius (PgmRenderShadowIFace *shadow, gint radius)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS
    (shadow)->set_radius (shadow, radius);
}

/**
 * pgm_render_shadow_get_radius:
 * @shadow: A #PgmRenderShadowIFace object.
 * @radius: The radius to fill.
 *
 * Retrieves the #radius from #shadow.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_get_radius (PgmRenderShadowIFace *shadow, gint *radius)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS
    (shadow)->get_radius (shadow, radius);
}

/**
 * pgm_render_shadow_set_offset:
 * @shadow: A #PgmRenderShadowIFace object.
 * @x: The X offset.
 * @y: The Y offset.
 *
 * Sets the offset of #shadow to (#x, #y).
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_set_offset (PgmRenderShadowIFace *shadow, gint x, gint y)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS (shadow)->set_offset (shadow, x, y);
}

/**
 * pgm_render_shadow_get_offset:
 * @shadow: A #PgmRenderShadowIFace object.
 * @x: The X offset to fill.
 * @y: The Y offset to fill.
 *
 * Retrieves the (#x, #y) offset from #shadow.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_get_offset (PgmRenderShadowIFace *shadow, gint *x, gint *y)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS (shadow)->get_offset (shadow, x, y);
}

/**
 * pgm_render_shadow_set_type:
 * @shadow: A #PgmRenderShadowIFace object.
 * @type: The type of the shadow.
 *
 * Sets the #type of #shadow.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_set_type (PgmRenderShadowIFace *shadow,
			    PgmRenderShadowType type)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS (shadow)->set_type (shadow, type);
}

/**
 * pgm_render_shadow_get_type:
 * @shadow: A #PgmRenderShadowIFace object.
 * @type: The type to fill.
 *
 * Retrieves the #type from #shadow.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_get_type (PgmRenderShadowIFace *shadow,
			    PgmRenderShadowType *type)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS (shadow)->get_type (shadow, type);
}

/**
 * pgm_render_shadow_set_color:
 * @shadow: A #PgmRenderShadowIFace object.
 * @r: The R color layer.
 * @g: The G color layer.
 * @b: The B color layer.
 * @a: The A color layer.
 *
 * Sets the color of #shadow to (#r, #g, #b, #a).
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_set_color (PgmRenderShadowIFace *shadow, guchar r, guchar g,
			  guchar b, guchar a)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS
    (shadow)->set_color (shadow, r, g, b, a);
}

/**
 * pgm_render_shadow_get_color:
 * @shadow: A #PgmRenderShadowIFace object.
 * @r: The R color layer to fill.
 * @g: The G color layer to fill.
 * @b: The B color layer to fill.
 * @a: The A color layer to fill.
 *
 * Retrieves the (#r, #g, #b, #a) color of #shadow.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_shadow_get_color (PgmRenderShadowIFace *shadow, guchar *r,
			     guchar *g, guchar *b, guchar *a)
{
  g_return_val_if_fail (PGM_RENDER_IS_SHADOW_IFACE (shadow),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_SHADOW_IFACE_GET_CLASS
    (shadow)->get_color (shadow, r, g, b, a);
}
