/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render OpenGL 1.2 plugin
 *
 * Copyright © 2007 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 *
 * Author(s): Colin Laplace <haiku at bloodshed.net>
 *            Loïc Molinari <loic@fluendo.com>
 */

#include "pgmrendergl1.h"
#include "pgmrendergl1window.h"
#include "pgmrender-glx.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <string.h>
#include <math.h>

#define USE_RANDR
#ifdef USE_RANDR
#  include <X11/extensions/Xrandr.h> 
#endif /* USE_RANDR */
#ifdef USE_VIDMOD
#  include <X11/extensions/xf86vmode.h>
#endif /* USE_VIDMOD */

static PgmRenderGl1Window *g_window = NULL;

/* FIXME: Duplication */
static void
reshape (PgmRenderGlProcAddressList *gl, gint canvas_width, gint canvas_height,
         gint window_width, gint window_height)
{
  gfloat fov = 0.15f;
  gfloat near_plane = 500.0f;
  gfloat far_plane = 5000.0f;
  gfloat tan_fov = tan (fov / (360.0f / G_PI));
  gfloat distance = 1.0f / tan_fov;
  gfloat n = (distance - near_plane) * tan_fov;

  gl->viewport (0, window_height-canvas_height, canvas_width, canvas_height);
  gl->matrix_mode (PGM_RENDER_GL_PROJECTION);
  gl->load_identity ();
  gl->frustum (-n, n, -n, n, distance-near_plane, distance+far_plane);
  gl->scale_f (2.0f/canvas_width, -2.0f/canvas_height, 1.0f);
  gl->translate_f (-canvas_width/2.0f, -canvas_height/2.0f, -distance);
  gl->matrix_mode (PGM_RENDER_GL_MODELVIEW);
}

static void
glx_wait_map_window (PgmRenderGlBackend *backend, Window win)
{
  XEvent ev;

  do 
    {
      XMaskEvent (backend->context->display, StructureNotifyMask, &ev);
    }
  while ((ev.type != MapNotify) || (ev.xmap.event != win));
}

static void
glx_wait_unmap_window (PgmRenderGlBackend *backend, Window win)
{
  XEvent ev;

  do 
    {
      XMaskEvent (backend->context->display, StructureNotifyMask, &ev);
    } 
  while ((ev.type != UnmapNotify) || (ev.xunmap.event != win));
}

static gboolean
glx_create_window (PgmRenderGlBackend *backend)
{
  gint  attr[] = { GLX_RGBA, GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE,
                   1, GLX_DOUBLEBUFFER, GLX_DEPTH_SIZE, 1, None };
  XSizeHints sizehints;
  Colormap color_map;
  Window root;
  gint win_x = 0;
  gint win_y = 0;
  backend->context->width = 800;
  backend->context->height = 600;

  backend->context->display = XOpenDisplay (NULL);
  if (!backend->context->display)
    {
      g_print ("Unable to open default display");
      return FALSE;
    }

  backend->context->screen = DefaultScreen (backend->context->display);
  backend->event_poll_fd = ConnectionNumber (backend->context->display);

  backend->context->vi = glXChooseVisual (backend->context->display, 
                                          backend->context->screen, attr);
  if (!backend->context->vi)
    {
      g_print ("Couldn't get an RGBA double-buffered visual\n");
      return FALSE;
    }

  root = DefaultRootWindow (backend->context->display);
  color_map = XCreateColormap (backend->context->display, 
                               root, backend->context->vi->visual, AllocNone);

  backend->context->swa.background_pixel = 0;
  backend->context->swa.border_pixel = 0;
  backend->context->swa.colormap = color_map;
  backend->context->swa.override_redirect = False;
  backend->context->swa.event_mask =
    ButtonPressMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask
    | PointerMotionMask | ResizeRedirectMask;

  /* Create the normal window */
  backend->context->win =
    XCreateWindow (backend->context->display,
                   root, win_x, win_y,
                   backend->context->width,  backend->context->height,
                   0, backend->context->vi->depth, InputOutput,
                   backend->context->vi->visual,
                   CWBackPixel | CWBorderPixel | CWColormap | CWEventMask,
                   &backend->context->swa);

  backend->context->fs_swa.background_pixel = 0;
  backend->context->fs_swa.border_pixel = 0;
  backend->context->fs_swa.colormap = color_map;
  backend->context->fs_swa.override_redirect = True;
  backend->context->fs_swa.event_mask =
    ButtonPressMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask
    | PointerMotionMask | StructureNotifyMask | SubstructureNotifyMask;

  /* Create the fullscreen window */
  backend->context->fs_win = 
    XCreateWindow (backend->context->display, root, win_x, win_y,
                   backend->context->width, backend->context->height,
                   0, backend->context->vi->depth, InputOutput,
                   backend->context->vi->visual,
                   CWColormap | CWEventMask | CWOverrideRedirect,
                   &backend->context->fs_swa);

   /* Set hints and properties */
  sizehints.x = win_x;
  sizehints.y = win_y;
  sizehints.width  = backend->context->width;
  sizehints.height = backend->context->height;
  sizehints.flags = USSize | USPosition;
  XSetNormalHints (backend->context->display,
                   backend->context->win, &sizehints);
  XSetStandardProperties (backend->context->display,
                          backend->context->win, "", "",
                          None, (char **) NULL, 0, &sizehints);

  XMapWindow (backend->context->display, backend->context->win);
  XGrabPointer (backend->context->display, backend->context->win, True, 0,
                GrabModeAsync, GrabModeAsync, backend->context->win, None,
                CurrentTime);

  XFlush (backend->context->display);

  return TRUE;
}

static gboolean
glx_destroy_window (PgmRenderGlBackend *backend)
{
  XCloseDisplay (backend->context->display);

  return TRUE;
}

static gboolean
glx_set_pixel_format (PgmRenderGlBackend *backend)
{
  return TRUE;
}

static gboolean
glx_create_context (PgmRenderGlBackend *backend)
{
  backend->context->context = glXCreateContext
    (backend->context->display, backend->context->vi, 0, TRUE);

  if (!backend->context->context)
    {
      g_print ("Cannot create OpenGL context\n");
      return FALSE;
    }

  /* FIXME: Must check the availablity of the GLX_SGI_video_sync extension */
/*   backend->context->extensions = */
/*     glXQueryExtensionsString (backend->context->display, */
/*                               backend->context->screen); */
/*   backend->context->get_video_sync = (GlxSgiGetVideoSyncFunc) */
/*     glXGetProcAddress ((const guchar *) "glXGetVideoSyncSGI"); */
/*   backend->context->wait_video_sync = (GlxSgiWaitVideoSyncFunc) */
/*     glXGetProcAddress ((const guchar *) "glXWaitVideoSyncSGI"); */

  return TRUE;
}

static gboolean
glx_destroy_context (PgmRenderGlBackend *backend)
{
  XDestroyWindow (backend->context->display, backend->context->win);
  glXDestroyContext (backend->context->display, backend->context->context);

  return TRUE;
}

static void
glx_get_display_data (PgmRenderGlBackend *backend)
{
  backend->resolution_width =
    DisplayWidth (backend->context->display, backend->context->screen);
  backend->resolution_height =
    DisplayHeight (backend->context->display, backend->context->screen);
  backend->screen_width_mm =
    DisplayWidthMM (backend->context->display, backend->context->screen);
  backend->screen_height_mm =
    DisplayHeightMM (backend->context->display, backend->context->screen);
}

static gboolean
glx_build_text_lists (PgmRenderGlBackend *backend)
{
  Display* display;
  XFontStruct* font_info;
  PgmRenderGlProcAddressList *gl = backend->proc_address;
  int first;
  int last;

  backend->context->text_lists = gl->gen_lists (256);
  if (!gl->is_list (backend->context->text_lists))
    {
      g_warning ("Unable to build text display lists\n");
      return FALSE;
    }

  display = XOpenDisplay (0);
  if (!display)
    {
      g_warning ("Unable to open X display\n");
      return FALSE;
    }

  font_info = XLoadQueryFont (display, "7x13bold");
  if (!font_info)
    {
      g_warning ("Unable to load X font \"7x13bold\"\n");
      font_info = XLoadQueryFont (display, "fixed");
      if (!font_info)
        {
          g_warning ("Unable to load X font \"fixed\"\n");
          XCloseDisplay (display);
          display = 0;
          return FALSE;
        }
    }

  first = font_info->min_char_or_byte2;
  last = font_info->max_char_or_byte2;

  /* FIXME: Broken on some systems, check for the font cache extension? */
  glXUseXFont (font_info->fid, first, last - first + 1,
               backend->context->text_lists + first);

  XCloseDisplay (display);
  display = 0;

  return TRUE;
}

static void
glx_destroy_text_lists (PgmRenderGlBackend *backend)
{
  backend->proc_address->delete_lists (backend->context->text_lists, 256);
}

static void
glx_raster_text (PgmRenderGlBackend *backend, const gchar *text, gint x,
                 gint y, gfloat r, gfloat g, gfloat b)
{
  PgmRenderGlProcAddressList *gl = backend->proc_address;

  gl->load_identity ();
  gl->disable (PGM_RENDER_GL_TEXTURE_2D);
  gl->disable (PGM_RENDER_GL_DEPTH_TEST);
  gl->push_attrib (PGM_RENDER_GL_LIST_BIT);
  gl->color_4f (r, g, b, 1.0f);
  gl->raster_pos_2f (x, y);
  gl->list_base (backend->context->text_lists);
  gl->call_lists (strlen (text), PGM_RENDER_GL_UNSIGNED_BYTE,
                  (PgmRenderGLubyte*) text);
  gl->pop_attrib ();
}

static gboolean
glx_make_current (PgmRenderGlBackend *backend)
{
  glXMakeCurrent (backend->context->display, backend->context->win,
                  backend->context->context);

  return TRUE;
}

static void
glx_swap_buffers (PgmRenderGlBackend *backend)
{
  if (backend->context->full_screen == TRUE)
    glXSwapBuffers (backend->context->display, backend->context->fs_win);
  else
    glXSwapBuffers (backend->context->display, backend->context->win);
}

static gpointer
glx_get_proc_address (PgmRenderGlBackend *backend, const gchar *proc_name)
{
  return glXGetProcAddress ((const guchar *) proc_name);
}

static void
glx_set_title (PgmRenderGlBackend *backend, const gchar *title,
               const gchar *icon)
{
  XClassHint hint;
  XTextProperty title_prop;

  XStoreName (backend->context->display, backend->context->win, title);

  hint.res_name = (gchar *) title;
  hint.res_class = (gchar *) title;
  XSetClassHint (backend->context->display, backend->context->win, &hint);

  if (XStringListToTextProperty ((gchar **) &title, 1, &title_prop))
    XSetWMName(backend->context->display, backend->context->win, &title_prop);

  XFlush (backend->context->display);
}

static void
glx_reset_resolution (PgmRenderGlBackend *backend)
{
#ifdef USE_RANDR
  gint nsizes;
  XRRScreenSize *sizes;
  XRRScreenConfiguration *screen_config;
  Rotation rotation;
  gint cur_s, cur_w, cur_h;

  /* Get the current screen configuration */
  screen_config = XRRGetScreenInfo
    (backend->context->display, DefaultRootWindow (backend->context->display));
  cur_s = XRRConfigCurrentConfiguration (screen_config, &rotation);

  /* Get the list of configuration sizes */
  sizes = XRRConfigSizes (screen_config, &nsizes);

  /* Get current configuration width and height */
  if ((cur_s >= 0) && (cur_s < nsizes))
    {
      cur_w = sizes[cur_s].width;
      cur_h = sizes[cur_s].height;
    }
  else
    {
      cur_w = 0;
      cur_h = 0;
    }

  /* Reset the screen configuration to the initial state, if needed */
  if ((backend->context->randr_size_id >= 0)
      && (backend->context->randr_size_id < nsizes)
      && ((sizes[backend->context->randr_size_id].width != cur_w)
          || (sizes[backend->context->randr_size_id].height != cur_h)))
    XRRSetScreenConfig (backend->context->display, screen_config, 
                        DefaultRootWindow (backend->context->display),
                        backend->context->randr_size_id, rotation,
                        CurrentTime);

  XRRFreeScreenConfigInfo (screen_config);
#endif /* USE_RANDR */

  XSync (backend->context->display, True);
}

static void
glx_set_resolution (PgmRenderGlBackend *backend, Display *dpy, int screen,
                    Window win, int w, int h)
{
#ifdef USE_RANDR
  gint i, nsizes;
  XRRScreenSize *sizes;
  XRRScreenConfiguration *screen_config;
  gint size_id;
  gint cur_w;
  gint cur_h;
  gint new_size_id = -1;
  Rotation rotation;

  screen_config = XRRGetScreenInfo (dpy, DefaultRootWindow (dpy));
  size_id = XRRConfigCurrentConfiguration (screen_config, &rotation);
  if (backend->context->randr_size_id == -1)
    backend->context->randr_size_id = size_id;
  sizes = XRRConfigSizes (screen_config, &nsizes);

  if ((size_id > 0) && (size_id < nsizes))
    {
      cur_w = sizes[size_id].width;
      cur_h = sizes[size_id].height;
    }
  else
    {
      cur_w = 0;
      cur_h = 0;
    }

  /* Find and set the config we want, if not already set */
  for (i = 0; i < nsizes; i++) 
    {
      if ((sizes[i].width == w)
          && (sizes[i].height == h)
          && ((sizes[i].width != cur_w)
              || (sizes[i].height != cur_h)))
        {
          new_size_id = i;
          break;
        }
    }
  /* If found, set it */
  if (new_size_id != -1)
    XRRSetScreenConfig (dpy, screen_config, DefaultRootWindow (dpy),
                        new_size_id, rotation, CurrentTime);

  XRRFreeScreenConfigInfo (screen_config);
#endif /* USE_RANDR */

#ifdef USE_VIDMOD
  int mode_count, i, found;
  XF86VidModeModeInfo **modes;
  
  found = -1;

  /* Get available video modes and sets one if matching */
  if (XF86VidModeGetAllModeLines (dpy, screen, &mode_count, &modes))
    {
      for (i = 0; i < mode_count; i++)
        {
          if ((modes[i]->hdisplay == w) && (modes[i]->vdisplay == h))
            {
              found = i;
              if (!XF86VidModeSwitchToMode (dpy, screen, modes[i]))
                g_warning ("Could not switch video mode\n");
              break;
            }
        }
    }

  if (found == -1)
    g_warning ("No suitable video modes found\n");
#endif /* USE_VIDMOD */
}

static gboolean
glx_set_size (PgmRenderGlBackend *backend, gint width, gint height)
{
  backend->context->width = width;
  backend->context->height = height;

  if (backend->context->full_screen)
    {
      glx_set_resolution (backend, backend->context->display,
                          backend->context->screen, backend->context->fs_win,
                          width, height);
      XResizeWindow (backend->context->display, backend->context->fs_win,
                     width, height);
    }
  else
    {
      backend->context->swa.override_redirect = True;
      XChangeWindowAttributes (backend->context->display,
                               backend->context->win, CWOverrideRedirect,
                               &backend->context->swa);
      XResizeWindow (backend->context->display, backend->context->win,
                     width, height);
      backend->context->swa.override_redirect = False;
      XChangeWindowAttributes (backend->context->display,
                               backend->context->win, CWOverrideRedirect,
                               &backend->context->swa);
      XResizeWindow (backend->context->display, backend->context->win,
                     width, height);
    }

  XSync (backend->context->display, True);

  return TRUE;
}

static gboolean
glx_set_fullscreen (PgmRenderGlBackend *backend, gboolean fullscreen,
                    gint width, gint height)
{
  PgmRenderEventConfigure *pgm_render_event;
  PgmRenderBaseWindow *window;

  backend->context->full_screen = fullscreen;
  backend->context->width = width;
  backend->context->height = height;

  if (fullscreen)
    {
      glXMakeCurrent (backend->context->display, backend->context->fs_win,
                      backend->context->context);
      XMoveResizeWindow (backend->context->display, backend->context->fs_win,
                         0, 0, backend->resolution_width,
                         backend->resolution_height);
      XUnmapWindow (backend->context->display, backend->context->win);
      XMapRaised (backend->context->display, backend->context->fs_win);
      glx_wait_map_window (backend, backend->context->fs_win);
      XRaiseWindow (backend->context->display, backend->context->fs_win);
      XGrabKeyboard (backend->context->display, backend->context->fs_win, True,
                    GrabModeAsync, GrabModeAsync, CurrentTime);
      XGrabPointer (backend->context->display, backend->context->fs_win, True,
                    0, GrabModeAsync, GrabModeAsync, backend->context->fs_win,
                    None, CurrentTime);
      glx_set_resolution (backend, backend->context->display,
                          backend->context->screen,
                          backend->context->fs_win, width, height);
    }
  else
    {
      glXMakeCurrent (backend->context->display, backend->context->win,
                      backend->context->context);
      glx_reset_resolution (backend);
      XUnmapWindow (backend->context->display, backend->context->fs_win);      
      glx_wait_unmap_window (backend, backend->context->fs_win);
      XMapRaised (backend->context->display, backend->context->win);      
      XRaiseWindow (backend->context->display, backend->context->win);
      XReparentWindow (backend->context->display, backend->context->win, 
                       DefaultRootWindow (backend->context->display), 0, 0);
    }

  /* Synthesize a Pigment event to notify the program */
  pgm_render_event = g_malloc (sizeof (PgmRenderEventConfigure));
  pgm_render_event->type = PGM_RENDER_CONFIGURE;
  pgm_render_event->time = 0;
  pgm_render_event->source = 0;
  pgm_render_event->width = width;
  pgm_render_event->height = height;
  g_window->event = g_list_prepend (g_window->event, pgm_render_event);

  /* FIXME */
  window = PGM_RENDER_BASE_WINDOW (g_window);
  if (window->scale_mode == PGM_RENDER_WINDOW_ADJUST
      || window->scale_mode == PGM_RENDER_WINDOW_SCREEN_ADJUST)
    g_window->new_size = TRUE;
  else
    {
      backend->set_size (g_window->backend, width, height);
      reshape (backend->proc_address, window->canvas_width,
               window->canvas_height, window->width, window->height);
    }

  XSync (backend->context->display, True);

  return TRUE;
}

static void
glx_wait_for_vblank (PgmRenderGlBackend *backend)
{
  guint retrace_count;
  backend->context->get_video_sync (&retrace_count);
  backend->context->wait_video_sync (2, (retrace_count + 1) % 2,
                                     &retrace_count);
}

static Cursor
get_none_cursor (PgmRenderGlBackend *backend, Window window)
{
  static Cursor none_cursor = None;
  gchar none_cursor_bits[32];
  Pixmap none_cursor_pixmap;
  XColor color;

  memset (none_cursor_bits, 0, sizeof (none_cursor_bits));
  memset (&color, 0, sizeof (color));

  none_cursor_pixmap = XCreateBitmapFromData (backend->context->display,
                                              window, none_cursor_bits, 16, 16);
  if (none_cursor_pixmap != None)
    {
      none_cursor = XCreatePixmapCursor (backend->context->display,
                                         none_cursor_pixmap,
                                         none_cursor_pixmap,
                                         &color, &color, 0, 0);
      XFreePixmap (backend->context->display, none_cursor_pixmap);
    }

  return none_cursor;
}

static gboolean
glx_set_cursor (PgmRenderGlBackend *backend, PgmRenderCursor cursor)
{
  Cursor cur;

  switch (cursor)
    {
    case PGM_RENDER_CURSOR_NONE:
      XDefineCursor (backend->context->display,  backend->context->win,
                     get_none_cursor (backend, backend->context->win));
      XDefineCursor (backend->context->display,  backend->context->fs_win,
                     get_none_cursor (backend, backend->context->fs_win));
      break;

    case PGM_RENDER_CURSOR_LEFT_ARROW:
      cur = XCreateFontCursor (backend->context->display, XC_top_left_arrow);
      XDefineCursor (backend->context->display,  backend->context->win, cur);
      XDefineCursor (backend->context->display,  backend->context->fs_win, cur);
      break;

    case PGM_RENDER_CURSOR_INHERIT:
      XDefineCursor (backend->context->display,  backend->context->win,
                     None);
      XDefineCursor (backend->context->display,  backend->context->fs_win,
                     None);
      break;

    default:
      break;
    }

  XSync (backend->context->display, True);

  return TRUE;
}

static gboolean
glx_check_events (PgmRenderGlBackend *backend)
{
  /* Do we have X events to process ? */
  return XPending (backend->context->display);
}

static void
set_key (guint *value, unsigned int key)
{
  switch (key)
    {
    case XK_Left:   *value = 20; break; 
    case XK_Up:     *value = 21; break;
    case XK_Right:  *value = 22; break;
    case XK_Down:   *value = 23; break;
    case XK_Return: *value = 18; break;
    case XK_Escape: *value = 19; break;
    case XK_space:  *value = 15; break;
    case XK_a:      *value = 50; break;
    case XK_b:      *value = 51; break;
    case XK_c:      *value = 52; break;
    case XK_d:      *value = 53; break;
    case XK_e:      *value = 54; break;
    case XK_f:      *value = 55; break;
    case XK_g:      *value = 56; break;
    case XK_h:      *value = 57; break;
    case XK_i:      *value = 58; break;
    case XK_j:      *value = 59; break;
    case XK_k:      *value = 60; break;
    case XK_l:      *value = 61; break;
    case XK_m:      *value = 62; break;
    case XK_n:      *value = 63; break;
    case XK_o:      *value = 64; break;
    case XK_p:      *value = 65; break;
    case XK_q:      *value = 66; break;
    case XK_r:      *value = 67; break;
    case XK_s:      *value = 68; break;
    case XK_t:      *value = 69; break;
    case XK_u:      *value = 70; break;
    case XK_v:      *value = 71; break;
    case XK_w:      *value = 72; break;
    case XK_x:      *value = 73; break;
    case XK_y:      *value = 74; break;
    case XK_z:      *value = 75; break;
    default:        *value = -1; break;
    }
}

static gboolean
glx_dispatch_events (PgmRenderGlBackend *backend)
{
  XEvent ev;

  /* Process an event if needed */
  if (XPending (backend->context->display))
    {
      XNextEvent (backend->context->display, &ev);
      
      switch (ev.type)
        {
        case KeyPress:
          {
            PgmRenderEventKey *pgm_render_event =
              g_malloc (sizeof (PgmRenderEventKey));
            pgm_render_event->type = PGM_RENDER_KEY_PRESS;
            pgm_render_event->time = 0;
            pgm_render_event->source = 0;
            set_key (&pgm_render_event->keyval, 
                     XKeycodeToKeysym (backend->context->display,
                                       ev.xkey.keycode, 0));
            g_window->event = g_list_prepend (g_window->event,
                                              pgm_render_event);
            break;          
          }
        case KeyRelease:
          {
            PgmRenderEventKey *pgm_render_event =
              g_malloc (sizeof (PgmRenderEventKey));
            pgm_render_event->type = PGM_RENDER_KEY_RELEASE;
            pgm_render_event->time = 0;
            pgm_render_event->source = 0;
            set_key (&pgm_render_event->keyval, 
                     XKeycodeToKeysym (backend->context->display,
                                       ev.xkey.keycode, 0));
            g_window->event = g_list_prepend (g_window->event,
                                              pgm_render_event);
            break;
          }
        case ButtonPress:
          {
            PgmRenderEventButton *pgm_render_event =
              g_malloc (sizeof (PgmRenderEventButton));
            pgm_render_event->type = PGM_RENDER_BUTTON_PRESS;
            pgm_render_event->time = 0;
            pgm_render_event->source = 0;
            pgm_render_event->x = ev.xbutton.x;
            pgm_render_event->y = ev.xbutton.y;
            g_window->event = g_list_prepend (g_window->event,
                                              pgm_render_event);
            break;
          }
        case ButtonRelease:
          {
            PgmRenderEventButton *pgm_render_event =
              g_malloc (sizeof (PgmRenderEventButton));
            pgm_render_event->type = PGM_RENDER_BUTTON_RELEASE;
            pgm_render_event->time = 0;
            pgm_render_event->source = 0;
            pgm_render_event->x = ev.xbutton.x;
            pgm_render_event->y = ev.xbutton.y;
            g_window->event = g_list_prepend (g_window->event,
                                              pgm_render_event);
            break;
          }
        case MotionNotify:
          {
            if ((ev.xmotion.x > 0)
                && (ev.xmotion.x < backend->context->width)
                && (ev.xmotion.y > 0)
                && (ev.xmotion.y < backend->context->height))
              {
                PgmRenderEventMotion *pgm_render_event =
                  g_malloc (sizeof (PgmRenderEventMotion));
                pgm_render_event->type = PGM_RENDER_MOTION_NOTIFY;
                pgm_render_event->time = 0;
                pgm_render_event->source = 0;
                pgm_render_event->x = ev.xmotion.x;
                pgm_render_event->y = ev.xmotion.y;
                g_window->event = g_list_prepend (g_window->event,
                                                  pgm_render_event);
              }
            break;
          }
        case ResizeRequest:
          {
            PgmRenderBaseWindow *window = PGM_RENDER_BASE_WINDOW (g_window);
            PgmRenderEventConfigure *pgm_render_event =
              g_malloc (sizeof (PgmRenderEventConfigure));
            pgm_render_event->type = PGM_RENDER_CONFIGURE;
            pgm_render_event->time = 0;
            pgm_render_event->source = 0;
            pgm_render_event->width = ev.xresizerequest.width;
            pgm_render_event->height = ev.xresizerequest.height;
            g_window->event = g_list_prepend (g_window->event,
                                              pgm_render_event);

            /* FIXME: Bad to do that here! */
            window->width = ev.xresizerequest.width;
            window->height = ev.xresizerequest.height;
            if (window->scale_mode == PGM_RENDER_WINDOW_ADJUST)
              g_window->new_size = TRUE;
            else if (window->scale_mode == PGM_RENDER_WINDOW_SCREEN_ADJUST)
              g_window->new_size = TRUE;
            else
              {
                backend->set_size (g_window->backend,
                                   ev.xresizerequest.width,
                                   ev.xresizerequest.height);
                reshape (backend->proc_address, window->canvas_width,
                         window->canvas_height, window->width,
                         window->height);
              }
            XSync (backend->context->display, True);
            break;
          }
        case DestroyNotify:
          {
            PgmRenderEventDelete *pgm_render_event =
              g_malloc (sizeof (PgmRenderEventDelete));
            pgm_render_event->type = PGM_RENDER_DELETE;
            pgm_render_event->time = 0;
            pgm_render_event->source = 0;
            g_window->event = g_list_prepend (g_window->event,
                                              pgm_render_event);
            break;
          }
        }
    }

  return TRUE;
}

void
glx_init_context (PgmRenderGl1Window *window)
{
  window->backend->create_window = glx_create_window;
  window->backend->set_title = glx_set_title;
  window->backend->destroy_window = glx_destroy_window;
  window->backend->set_pixel_format = glx_set_pixel_format;
  window->backend->create_context = glx_create_context;
  window->backend->destroy_context = glx_destroy_context;
  window->backend->make_current = glx_make_current;
  window->backend->swap_buffers = glx_swap_buffers;
  window->backend->get_proc_address = glx_get_proc_address;
  window->backend->check_events = glx_check_events;
  window->backend->dispatch_events = glx_dispatch_events;
  window->backend->set_size = glx_set_size;
  window->backend->set_fullscreen = glx_set_fullscreen;
  window->backend->get_display_data = glx_get_display_data;
  window->backend->build_text_lists = glx_build_text_lists;
  window->backend->destroy_text_lists = glx_destroy_text_lists;
  window->backend->raster_text = glx_raster_text;
  window->backend->wait_for_vblank = glx_wait_for_vblank;
  window->backend->set_cursor = glx_set_cursor;
  window->backend->event_poll_fd = 0;
  g_window = window;

  window->backend->context = g_new0 (PgmRenderGlxContext, 1);
  window->backend->context->randr_size_id = -1;
  window->backend->context->full_screen = FALSE;
}

void
glx_fini_context (PgmRenderGl1Window *window)
{
  window->backend->create_window = NULL;
  window->backend->set_title = NULL;
  window->backend->destroy_window = NULL;
  window->backend->set_pixel_format = NULL;
  window->backend->create_context = NULL;
  window->backend->destroy_context = NULL;
  window->backend->make_current = NULL;
  window->backend->swap_buffers = NULL;
  window->backend->get_proc_address = NULL;
  window->backend->check_events = NULL;
  window->backend->dispatch_events = NULL;
  window->backend->set_size = NULL;
  window->backend->set_fullscreen = NULL;
  window->backend->get_display_data = NULL;
  window->backend->build_text_lists = NULL;
  window->backend->destroy_text_lists = NULL;
  window->backend->raster_text = NULL;
  window->backend->wait_for_vblank = NULL;
  window->backend->set_cursor = NULL;
  window->backend->event_poll_fd = 0;
  g_window = NULL;

  g_free (window->backend->context);
}
