/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render OpenGL 1.2 plugin
 *
 * Copyright © 2006, 2007 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 *
 * The OpenGL extensions loading code have been inspired by the Glitz
 * image compositing library <www.freedesktop.org/wiki/Software/glitz>.
 * Copyright © 2004, David Reveman, Peter Nilsson
 *
 * Author: Loïc Molinari <loic@fluendo.com>
 */

#include "pgmrendergl1window.h"
#include "pgmrendergl1.h"
#include <GL/gl.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

static void pgm_render_gl1_window_class_init (PgmRenderGl1WindowClass *klass);
static void pgm_render_gl1_window_init (PgmRenderGl1Window *obj);
static void pgm_render_gl1_window_finalize (GObject *obj);

static PgmRenderError pgm_render_gl1_window_update (PgmRenderBaseWindow *window);
static PgmRenderError pgm_render_gl1_window_set_cursor (PgmRenderBaseWindow *window, PgmRenderCursor cursor);
static PgmRenderError pgm_render_gl1_window_set_color (PgmRenderBaseWindow *window, guchar r, guchar g, guchar b, guchar a);
static PgmRenderError pgm_render_gl1_window_get_color (PgmRenderBaseWindow *window, guchar *r, guchar *g, guchar *b, guchar *a);
static PgmRenderError pgm_render_gl1_window_set_title (PgmRenderBaseWindow *window, const gchar *window_title, const gchar *icon_title);
static PgmRenderError pgm_render_gl1_window_set_size (PgmRenderBaseWindow *window, gint width, gint height);
static PgmRenderError pgm_render_gl1_window_set_canvas_size (PgmRenderBaseWindow *window, gint canvas_width, gint canvas_height);
static PgmRenderError pgm_render_gl1_window_get_events (PgmRenderBaseWindow *window, GList **event);
static PgmRenderError pgm_render_gl1_window_start_loop (PgmRenderBaseWindow *window);
static PgmRenderError pgm_render_gl1_window_quit_loop (PgmRenderBaseWindow *window);
static PgmRenderError pgm_render_gl1_window_add_drawable (PgmRenderBaseWindow *window, PgmRenderDrawable *drawable, PgmRenderDrawableOrder order);
static PgmRenderError pgm_render_gl1_window_del_drawable (PgmRenderBaseWindow *window, PgmRenderDrawable *drawable);
static PgmRenderError pgm_render_gl1_window_reorder_drawable (PgmRenderBaseWindow *window, PgmRenderDrawable *drawable, gint position);
static PgmRenderError pgm_render_gl1_window_regenerate_drawables (PgmRenderBaseWindow *window);
static PgmRenderError pgm_render_gl1_window_set_fullscreen (PgmRenderBaseWindow *window, gboolean fullscreen, gint width, gint height);
static PgmRenderError pgm_render_gl1_window_get_fullscreen (PgmRenderBaseWindow *window, gboolean *fullscreen);
static PgmRenderError pgm_render_gl1_window_set_scale (PgmRenderBaseWindow *window, gfloat sx, gfloat sy);
static PgmRenderError pgm_render_gl1_window_set_scale_mode (PgmRenderBaseWindow *window, PgmRenderWindowScaleMode scale_mode);
static PgmRenderError pgm_render_gl1_window_get_screen_resolution (PgmRenderBaseWindow *window, gint *width, gint *height);
static PgmRenderError pgm_render_gl1_window_get_screen_size (PgmRenderBaseWindow *window, gint *width, gint *height);
static PgmRenderError pgm_render_gl1_window_get_screen_size_mm (PgmRenderBaseWindow *window, gint *width, gint *height);

static gboolean render (gpointer data);
static gpointer render_loop (gpointer data);

static gboolean pgm_render_gl1_event_prepare (GSource *source, gint *timeout);
static gboolean pgm_render_gl1_event_check (GSource *source);
static gboolean pgm_render_gl1_event_dispatch (GSource *source, GSourceFunc callback, gpointer user_data);

static void _set_scale (PgmRenderGl1Window *window, gfloat sx, gfloat sy);

static PgmRenderBaseWindowClass *parent_class = NULL;

static GSourceFuncs pgm_render_gl1_event_funcs = {
  pgm_render_gl1_event_prepare,
  pgm_render_gl1_event_check,
  pgm_render_gl1_event_dispatch,
  NULL
};

G_DEFINE_TYPE (PgmRenderGl1Window, pgm_render_gl1_window,
               PGM_RENDER_TYPE_BASE_WINDOW);

/* Global pointer to the Window to use in the source event */
static PgmRenderGl1Window *g_window = NULL;

/* GLSL vertex shaders */

/* Main function */
static const gchar *vertex_main[] =
  {
    "varying vec4 blend_color;"

    "void main (void)",
    "{",
    "  blend_color = gl_Color;"
    "  gl_TexCoord[0] = gl_MultiTexCoord0;",
    "  gl_Position = ftransform ();",
    "}",
    NULL
  };

/* GLSL fragment shaders */
/* FIXME: Still need optimisations. */

/* YUV to RGBA color space conversions header */
static const gchar *fragment_header[] =
  {
    "uniform sampler2D tex0, tex1, tex2;",
    "uniform float tex_width, tex_height;",
    "/* uniform mat4 filter; */",
    "varying vec4 blend_color;",

    /* ITU-R BT.601 color standard for SDTV */
    "const vec3 BT601_bias = vec3 (-0.0625, -0.5, -0.5);",
    "const mat3 BT601 = mat3 (1.1640,  0.0000,  1.4022,",
    "                         1.1640, -0.3910, -0.8130,",
    "                         1.1640,  2.0180,  0.0000);",

    /* ITU-R BT.709 color standard for HDTV */
    "const vec3 BT709_bias = vec3 (0.0, -0.5, -0.5);",
    "const mat3 BT709 = mat3 (1.0000,  0.0000,  1.4022,",
    "                         1.0000, -0.3456, -0.7145,",
    "                         1.0000,  1.7710,  0.0000);",

    /* Convenient constants */
    "const int Y = 0;",
    "const int Cb = 1;",
    "const int Cr = 2;",
    NULL
  };

/* RGB color space: RGB.
 * OpenGL internal format: GL_RGB. */
static const gchar *rgb_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 rgb_nearest_2d (in vec2 coord)",
    "{",
    "  return texture2D (tex0, coord);",
    "}",

    /* Bilinear filtering 2d texture sampler */
/*     "vec4 rgb_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */
/*     "  vec2 position;", */

/*     "  texel_size_x = 1.0 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (coord.s * tex_width),", */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = rgb_nearest_2d (coord);", */
/*     "  color_10 = rgb_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = rgb_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = rgb_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "              mix (color_01, color_11, position.s),", */
/*     "              position.t);", */
/*     "}", */
    NULL
  };

/* RGB color space: BGR.
 * OpenGL internal format: GL_RGB. */
static const gchar *bgr_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 bgr_nearest_2d (in vec2 coord)",
    "{",
    "  return texture2D (tex0, coord).bgra;",
    "}",

    /* Bilinear filtering 2d texture sampler */
/*     "vec4 bgr_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */
/*     "  vec2 position;", */

/*     "  texel_size_x = 1.0 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (coord.s * tex_width),", */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = bgr_nearest_2d (coord);", */
/*     "  color_10 = bgr_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = bgr_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = bgr_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "              mix (color_01, color_11, position.s),", */
/*     "              position.t);", */
/*     "}", */
    NULL
  };

/* RGB color space: RGBA.
 * OpenGL internal format: GL_RGBA. */
static const gchar *rgba_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 rgba_nearest_2d (in vec2 coord)",
    "{",
    "  return texture2D (tex0, coord);",
    "}",

    /* Bilinear filtering 2d texture sampler */
/*     "vec4 rgba_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */
/*     "  vec2 position;", */

/*     "  texel_size_x = 1.0 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (coord.s * tex_width),", */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = rgba_nearest_2d (coord);", */
/*     "  color_10 = rgba_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = rgba_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = rgba_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "              mix (color_01, color_11, position.s),", */
/*     "              position.t);", */
/*     "}", */
    NULL
  };

/* RGB color space: BGRA.
 * OpenGL internal format: GL_RGBA. */
static const gchar *bgra_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 bgra_nearest_2d (in vec2 coord)",
    "{",
    "  return texture2D (tex0, coord);",
    "}",

    /* Bilinear filtering 2d texture sampler */
/*     "vec4 bgra_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */
/*     "  vec2 position;", */

/*     "  texel_size_x = 1.0 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (coord.s * tex_width),", */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = bgra_nearest_2d (coord);", */
/*     "  color_10 = bgra_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = bgra_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = bgra_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "              mix (color_01, color_11, position.s),", */
/*     "              position.t);", */
/*     "}", */
    NULL
  };

/* Planar FOURCC formats: I420, IYUV.
 * OpenGL internal format: GL_LUMINANCE. */
static const gchar *i420_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 i420_nearest_2d (in vec2 coord)",
    "{",
    "  vec3 color;",
    
    "  color[Y] = texture2D (tex0, coord).r + BT601_bias.x;",
    "  color[Cb] = texture2D (tex1, coord).r + BT601_bias.y;",
    "  color[Cr] = texture2D (tex2, coord).r + BT601_bias.z;",

    "  return vec4 (color * BT601, 1.0);",
    "}",

    /* Bilinear filtering 2d texture sampler */
/*     "vec4 i420_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */
/*     "  vec2 position;", */

/*     "  texel_size_x = 1.0 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (coord.s * tex_width),", */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = i420_nearest_2d (coord);", */
/*     "  color_10 = i420_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = i420_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = i420_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "              mix (color_01, color_11, position.s),", */
/*     "              position.t);", */
/*     "}", */
    NULL
  };

/* Planar FOURCC format: YV12.
 * OpenGL internal format: GL_LUMINANCE. */
static const gchar *yv12_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 yv12_nearest_2d (in vec2 coord)",
    "{",
    "  vec3 color;",

    "  color[Y] = texture2D (tex0, coord).r + BT601_bias.x;",
    "  color[Cb] = texture2D (tex2, coord).r + BT601_bias.y;",
    "  color[Cr] = texture2D (tex1, coord).r + BT601_bias.z;",

    "  return vec4 (color * BT601, 1.0);",
    "}",

    /* Bilinear filtering 2d texture sampler */
 /*    "vec4 yv12_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */
/*     "  vec2 position;", */

/*     "  texel_size_x = 1.0 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (coord.s * tex_width),", */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = yv12_nearest_2d (coord);", */
/*     "  color_10 = yv12_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = yv12_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = yv12_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "              mix (color_01, color_11, position.s),", */
/*     "              position.t);", */
/*     "}", */
    NULL
  };

/* Packed FOURCC formats: YUYV, YUY2, V422, YUNV.
 * OpenGL internal format: GL_RGBA. */
static const gchar *yuyv_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 yuyv_nearest_2d (in vec2 coord)",
    "{",
    "  vec4 macropixel;",
    "  vec3 color;",

    "  macropixel = texture2D (tex0, coord);",
    "  color[Y] = (fract (coord.s * tex_width) < 0.5 ?"
    "              macropixel.r : macropixel.b) + BT601_bias.x;",
    "  color[Cb] = macropixel.g + BT601_bias.y;",
    "  color[Cr] = macropixel.a + BT601_bias.z;",

    "  return vec4 (color * BT601, 1.0);",
    "}",

    /* Bilinear filtering 2d texture sampler */
/*     "vec4 yuyv_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  vec2 position;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */

/*     "  texel_size_x = 0.5 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (2.0 * coord.s * tex_width)," */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = yuyv_nearest_2d (coord);", */
/*     "  color_10 = yuyv_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = yuyv_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = yuyv_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "              mix (color_01, color_11, position.s),", */
/*     "              position.t);", */
/*     "}", */
    NULL
  };

/* Packed FOURCC formats: UYVY, Y422, UYNV.
 * OpenGL internal format: GL_RGBA. */
static const gchar *uyvy_texture_samplers[] =
  {
    /* Nearest filtering 2d texture sampler */
    "vec4 uyvy_nearest_2d (in vec2 coord)",
    "{",
    "  vec4 macropixel;",
    "  vec3 color;",

    "  macropixel = texture2D (tex0, coord);",
    "  color[Y] = (fract (coord.s * tex_width) < 0.5 ?"
    "              macropixel.g : macropixel.a) + BT601_bias.x;",
    "  color[Cb] = macropixel.r + BT601_bias.y;",
    "  color[Cr] = macropixel.b + BT601_bias.z;",

    "  return vec4 (color * BT601, 1.0);",
    "}",

    /* Bilinear filtering 2d texture sampler */
/*     "vec4 uyvy_bilinear_2d (in vec2 coord)", */
/*     "{", */
/*     "  vec4 color_00, color_01, color_10, color_11;", */
/*     "  vec2 position;", */
/*     "  float texel_size_x, texel_size_y, s1, t1;", */

/*     "  texel_size_x = 0.5 / tex_width;", */
/*     "  texel_size_y = 1.0 / tex_height;", */
/*     "  coord.s -= 0.5 * texel_size_x;", */
/*     "  coord.t -= 0.5 * texel_size_y;", */
/*     "  position = vec2 (fract (2.0 * coord.s * tex_width)," */
/*     "                   fract (coord.t * tex_height));", */
/*     "  s1 = coord.s + texel_size_x;", */
/*     "  t1 = coord.t + texel_size_y;", */
/*     "  color_00 = uyvy_nearest_2d (coord);", */
/*     "  color_10 = uyvy_nearest_2d (vec2 (s1, coord.t));", */
/*     "  color_01 = uyvy_nearest_2d (vec2 (coord.s, t1));", */
/*     "  color_11 = uyvy_nearest_2d (vec2 (s1, t1));", */

/*     "  return mix (mix (color_00, color_10, position.s),", */
/*     "	      mix (color_01, color_11, position.s),", */
/*     "	      position.t);", */
/*     "}", */
    NULL
  };

/* Main function */
static const gchar fragment_main[] =
  "void main (void)"
  "{"
  "  gl_FragColor = %s (gl_TexCoord[0].st) * blend_color;"
  "}";

/* OpenGL extensions to be checked at runtime */
static PgmRenderGlExtensionMap opengl_extensions_map[] = {
  { "GL_ARB_texture_rectangle",
    PGM_RENDER_GL_FEATURE_TEXTURE_RECTANGLE },
  { "GL_EXT_texture_rectangle",
    PGM_RENDER_GL_FEATURE_TEXTURE_RECTANGLE },
  { "GL_NV_texture_rectangle",
    PGM_RENDER_GL_FEATURE_TEXTURE_RECTANGLE },
  { "GL_ARB_texture_non_power_of_two",
    PGM_RENDER_GL_FEATURE_TEXTURE_NON_POWER_OF_TWO },
  { "GL_ARB_texture_mirrored_repeat",
    PGM_RENDER_GL_FEATURE_TEXTURE_MIRRORED_REPEAT },
  { "GL_ARB_texture_border_clamp",
    PGM_RENDER_GL_FEATURE_TEXTURE_BORDER_CLAMP },
  { "GL_ARB_texture_env_combine",
    PGM_RENDER_GL_FEATURE_TEXTURE_ENV_COMBINE },
  { "GL_EXT_texture_env_combine",
    PGM_RENDER_GL_FEATURE_TEXTURE_ENV_COMBINE },
  { "GL_ARB_texture_env_dot3",
    PGM_RENDER_GL_FEATURE_TEXTURE_ENV_DOT3 },
  { "GL_ARB_multisample",
    PGM_RENDER_GL_FEATURE_MULTISAMPLE },
  { "GL_NV_multisample_filter_hint",
    PGM_RENDER_GL_FEATURE_MULTISAMPLE_FILTER_HINT },
  { "GL_ARB_multitexture",
    PGM_RENDER_GL_FEATURE_MULTITEXTURE },
  { "GL_EXT_multi_draw_arrays",
    PGM_RENDER_GL_FEATURE_MULTI_DRAW_ARRAYS },
  { "GL_ARB_fragment_program",
    PGM_RENDER_GL_FEATURE_FRAGMENT_PROGRAM },
  { "GL_ARB_shading_language_100",
    PGM_RENDER_GL_FEATURE_SHADING_LANGUAGE },
  { "GL_ARB_shader_objects",
    PGM_RENDER_GL_FEATURE_SHADER_OBJECTS },
  { "GL_ARB_fragment_shader",
    PGM_RENDER_GL_FEATURE_FRAGMENT_SHADER },
  { "GL_ARB_vertex_shader",
    PGM_RENDER_GL_FEATURE_VERTEX_SHADER },
  { "GL_ARB_vertex_buffer_object",
    PGM_RENDER_GL_FEATURE_VERTEX_BUFFER_OBJECT },
  { "GL_ARB_pixel_buffer_object",
    PGM_RENDER_GL_FEATURE_PIXEL_BUFFER_OBJECT },
  { "GL_EXT_pixel_buffer_object",
    PGM_RENDER_GL_FEATURE_PIXEL_BUFFER_OBJECT },
  { "GL_EXT_blend_color",
    PGM_RENDER_GL_FEATURE_BLEND_COLOR },
  { "GL_ARB_imaging",
    PGM_RENDER_GL_FEATURE_BLEND_COLOR },
  { "GL_APPLE_packed_pixels",
    PGM_RENDER_GL_FEATURE_PACKED_PIXELS },
  { "GL_EXT_framebuffer_object",
    PGM_RENDER_GL_FEATURE_FRAMEBUFFER_OBJECT },
  { NULL, 0 }
};

static PgmRenderGlProcAddressList opengl_proc_address_list = {
  /* OpenGL 1.2 core functions */
  (pgm_render_gl_enable) glEnable,
  (pgm_render_gl_disable) glDisable,
  (pgm_render_gl_get_error) glGetError,
  (pgm_render_gl_get_string) glGetString,
  (pgm_render_gl_begin) glBegin,
  (pgm_render_gl_end) glEnd,
  (pgm_render_gl_vertex_3f) glVertex3f,
  (pgm_render_gl_tex_coord_2f) glTexCoord2f,
  (pgm_render_gl_enable_client_state) glEnableClientState,
  (pgm_render_gl_disable_client_state) glDisableClientState,
  (pgm_render_gl_vertex_pointer) glVertexPointer,
  (pgm_render_gl_tex_coord_pointer) glTexCoordPointer,
  (pgm_render_gl_draw_arrays) glDrawArrays,
  (pgm_render_gl_tex_env_f) glTexEnvf,
  (pgm_render_gl_tex_env_fv) glTexEnvfv,
  (pgm_render_gl_tex_gen_i) glTexGeni,
  (pgm_render_gl_tex_gen_fv) glTexGenfv,
  (pgm_render_gl_color_4us) glColor4us,
  (pgm_render_gl_color_4f) glColor4f,
  (pgm_render_gl_scissor) glScissor,
  (pgm_render_gl_blend_func) glBlendFunc,
  (pgm_render_gl_line_width) glLineWidth,
  (pgm_render_gl_clear) glClear,
  (pgm_render_gl_clear_color) glClearColor,
  (pgm_render_gl_clear_stencil) glClearStencil,
  (pgm_render_gl_stencil_func) glStencilFunc,
  (pgm_render_gl_stencil_op) glStencilOp,
  (pgm_render_gl_push_attrib) glPushAttrib,
  (pgm_render_gl_pop_attrib) glPopAttrib,
  (pgm_render_gl_matrix_mode) glMatrixMode,
  (pgm_render_gl_push_matrix) glPushMatrix,
  (pgm_render_gl_pop_matrix) glPopMatrix,
  (pgm_render_gl_load_identity) glLoadIdentity,
  (pgm_render_gl_load_matrix_f) glLoadMatrixf,
  (pgm_render_gl_depth_range) glDepthRange,
  (pgm_render_gl_viewport) glViewport,
  (pgm_render_gl_raster_pos_2f) glRasterPos2f,
  (pgm_render_gl_bitmap) glBitmap,
  (pgm_render_gl_read_buffer) glReadBuffer,
  (pgm_render_gl_draw_buffer) glDrawBuffer,
  (pgm_render_gl_copy_pixels) glCopyPixels,
  (pgm_render_gl_flush) glFlush,
  (pgm_render_gl_finish) glFinish,
  (pgm_render_gl_pixel_store_i) glPixelStorei,
  (pgm_render_gl_frustum) glFrustum,
  (pgm_render_gl_ortho) glOrtho,
  (pgm_render_gl_scale_f) glScalef,
  (pgm_render_gl_translate_f) glTranslatef,
  (pgm_render_gl_rotate_f) glRotatef,
  (pgm_render_gl_is_list) glIsList,
  (pgm_render_gl_delete_lists) glDeleteLists,
  (pgm_render_gl_gen_lists) glGenLists,
  (pgm_render_gl_new_list) glNewList,
  (pgm_render_gl_end_list) glEndList,
  (pgm_render_gl_call_list) glCallList,
  (pgm_render_gl_call_lists) glCallLists,
  (pgm_render_gl_list_base) glListBase,
  (pgm_render_gl_hint) glHint,
  (pgm_render_gl_depth_mask) glDepthMask,
  (pgm_render_gl_polygon_mode) glPolygonMode,
  (pgm_render_gl_shade_model) glShadeModel,
  (pgm_render_gl_color_mask) glColorMask,
  (pgm_render_gl_read_pixels) glReadPixels,
  (pgm_render_gl_get_tex_image) glGetTexImage,
  (pgm_render_gl_tex_sub_image_2d) glTexSubImage2D,
  (pgm_render_gl_gen_textures) glGenTextures,
  (pgm_render_gl_delete_textures) glDeleteTextures,
  (pgm_render_gl_bind_texture) glBindTexture,
  (pgm_render_gl_tex_image_2d) glTexImage2D,
  (pgm_render_gl_tex_parameter_i) glTexParameteri,
  (pgm_render_gl_tex_parameter_fv) glTexParameterfv,
  (pgm_render_gl_get_tex_level_parameter_iv) glGetTexLevelParameteriv,
  (pgm_render_gl_copy_tex_sub_image_2d) glCopyTexSubImage2D,
  (pgm_render_gl_get_integer_v) glGetIntegerv,
  (pgm_render_gl_get_float_v) glGetFloatv,
  /* OpenGL 1.2 extension functions */
  (pgm_render_gl_blend_color) 0,
  (pgm_render_gl_active_texture) 0,
  (pgm_render_gl_client_active_texture) 0,
  (pgm_render_gl_multi_draw_arrays) 0,
  (pgm_render_gl_gen_programs) 0,
  (pgm_render_gl_delete_programs) 0,
  (pgm_render_gl_program_string) 0,
  (pgm_render_gl_bind_program) 0,
  (pgm_render_gl_program_local_param_4fv) 0,
  (pgm_render_gl_get_program_iv) 0,
  (pgm_render_gl_gen_buffers) 0,
  (pgm_render_gl_delete_buffers) 0,
  (pgm_render_gl_bind_buffer) 0,
  (pgm_render_gl_buffer_data) 0,
  (pgm_render_gl_buffer_sub_data) 0,
  (pgm_render_gl_get_buffer_sub_data) 0,
  (pgm_render_gl_map_buffer) 0,
  (pgm_render_gl_unmap_buffer) 0,
  (pgm_render_gl_gen_framebuffers) 0,
  (pgm_render_gl_delete_framebuffers) 0,
  (pgm_render_gl_bind_framebuffer) 0,
  (pgm_render_gl_framebuffer_renderbuffer) 0,
  (pgm_render_gl_framebuffer_texture_2d) 0,
  (pgm_render_gl_check_framebuffer_status) 0,
  (pgm_render_gl_gen_renderbuffers) 0,
  (pgm_render_gl_delete_renderbuffers) 0,
  (pgm_render_gl_bind_renderbuffer) 0,
  (pgm_render_gl_renderbuffer_storage) 0,
  (pgm_render_gl_get_renderbuffer_parameter_iv) 0,
  (pgm_render_gl_delete_object) 0,
  (pgm_render_gl_get_handle) 0,
  (pgm_render_gl_detach_object) 0,
  (pgm_render_gl_create_shader_object) 0,
  (pgm_render_gl_shader_source) 0,
  (pgm_render_gl_compile_shader) 0,
  (pgm_render_gl_create_program_object) 0,
  (pgm_render_gl_attach_object) 0,
  (pgm_render_gl_link_program) 0,
  (pgm_render_gl_use_program_object) 0,
  (pgm_render_gl_get_object_parameter_iv) 0,
  (pgm_render_gl_get_info_log) 0,
  (pgm_render_gl_get_uniform_location) 0,
  (pgm_render_gl_uniform_matrix_4fv) 0,
  (pgm_render_gl_uniform_1f) 0,
  (pgm_render_gl_uniform_1i) 0,
};

static gboolean
check_opengl_extension (const gchar *extensions,
                        const gchar *extension_name)
{
  gchar *end;
  gint extension_name_len = strlen (extension_name);
  gchar *p = (gchar*) extensions;
  size_t size;

  if (!extensions)
    return FALSE;

  end = p + strlen (p);

  while (p < end)
    {
      size = strcspn (p, " ");

      if ((extension_name_len == size)
          && (strncmp (extension_name, p, size) == 0))
        return TRUE;

      p += size + 1;
    }

  return FALSE;
}

static gfloat
opengl_version_string_to_float (const gchar *version_string)
{
  gfloat version = 0.0f;
  gint i;

  for (i = 0; isdigit (version_string[i]); i++)
    version = version * 10.0f + (version_string[i] - 48);

  if (version_string[i++] != '.')
    return 0.0f;

  version = version * 10.0f + (version_string[i] - 48);

  return (version + 0.1f) / 10.0f;
}

static gint
get_shader_code_line_size (const gchar **shader_code)
{
  gint line = 0;

  while (shader_code[line])
    line++;

  return line;
}

static gchar **
concatenate_sampler_code (const gchar **header, const gchar **samplers,
                          const gchar *main, const gchar *sampler_name)
{
  gchar **shader_code;
  gint nb_lines, shader_code_line, line;

  nb_lines = get_shader_code_line_size (header)
    + get_shader_code_line_size (samplers) + 2;

  shader_code = (gchar**) g_malloc (nb_lines * sizeof (gchar*));

  shader_code_line = 0;
  while (header[shader_code_line])
    {
      shader_code[shader_code_line] = g_strdup (header[shader_code_line]);
      shader_code_line++;
    }
  line = 0;
  while (samplers[line])
    {
      shader_code[shader_code_line] = g_strdup (samplers[line]);
      shader_code_line++;
      line++;
    }
  shader_code[shader_code_line] = g_strdup_printf (main, sampler_name);
  shader_code[shader_code_line+1] = NULL;

  return shader_code;  
}

static void
free_shader_code (gchar **shader_code)
{
  gint line = 0;

  while (shader_code[line])
    {
      g_free (shader_code[line]);
      line++;
    }

  g_free (shader_code);
}

static gchar *
get_glsl_info_log (PgmRenderGLhandle object,
                   PgmRenderGlProcAddressList *gl)
{
  gint length = 0;
  gint written = 0;
  gchar *info_log;

  gl->get_object_parameter_iv
    (object, PGM_RENDER_GL_OBJECT_INFO_LOG_LENGTH, &length);

  if (length > 0)
    {
      info_log = (gchar *) g_malloc (length * sizeof (gchar));
      gl->get_info_log (object, length, &written, info_log);
      return info_log;
    }
  else
    return NULL;
}

/* FIXME: check/report/handle compilation and linking errors */
static gint
build_color_space_shaders (PgmRenderGl1Window *window)
{
  FragmentCode code[NB_COLOR_SPACES] =
    {
      { rgb_texture_samplers, "rgb_nearest_2d", "rgb_bilinear_2d" },
      { bgr_texture_samplers, "bgr_nearest_2d", "bgr_bilinear_2d" },
      { rgba_texture_samplers, "rgba_nearest_2d", "rgba_bilinear_2d" },
      { bgra_texture_samplers, "bgra_nearest_2d", "bgra_bilinear_2d" },
      { i420_texture_samplers, "i420_nearest_2d", "i420_bilinear_2d" },
      { yv12_texture_samplers, "yv12_nearest_2d", "yv12_bilinear_2d" },
      { uyvy_texture_samplers, "uyvy_nearest_2d", "uyvy_bilinear_2d" },
      { yuyv_texture_samplers, "yuyv_nearest_2d", "yuyv_bilinear_2d" }
    };
  PgmRenderGlProcAddressList *gl = window->backend->proc_address;
  PgmRenderGLhandle vertex_shader, fragment_shader;
  ColorSpaceShader shader;
  gchar **fragment_code;
  gchar *log;
  gint i;

  /* Vertex shader */
  vertex_shader = gl->create_shader_object (PGM_RENDER_GL_VERTEX_SHADER);
  gl->shader_source (vertex_shader, get_shader_code_line_size (vertex_main),
                     vertex_main, NULL);
  gl->compile_shader (vertex_shader);
  log = get_glsl_info_log (vertex_shader, gl);
  if (log)
    {
/*       g_print ("\"generic\" vertex shader compilation: \"%s\"\n", log); */
      g_free (log);
    }

  /* Fragment shaders */
  for (i = 0; i < NB_COLOR_SPACES; i++)
    {
      /* Nearest sampler */
      fragment_code = concatenate_sampler_code (fragment_header,
                                                code[i].sampler,
                                                fragment_main,
                                                code[i].nearest_func_name);
      fragment_shader = gl->create_shader_object (PGM_RENDER_GL_FRAGMENT_SHADER);
      gl->shader_source (fragment_shader,
                         get_shader_code_line_size
                         ((const gchar**) fragment_code),
                         (const gchar**) fragment_code, NULL);
      free_shader_code (fragment_code);
      gl->compile_shader (fragment_shader);
      log = get_glsl_info_log (vertex_shader, gl);
      if (log)
        {
/*           g_print ("\"%s\" fragment shader compilation: \"%s\"\n", */
/*                    code[i].nearest_func_name, log); */
          g_free (log);
        }

      window->color_space_shader[i][0].program = gl->create_program_object ();
      shader = window->color_space_shader[i][0];
      gl->attach_object (shader.program, vertex_shader);
      gl->attach_object (shader.program, fragment_shader);
      gl->link_program (shader.program);
      log = get_glsl_info_log (shader.program, gl);
      if (log)
        {
/*           g_print ("\"%s\" program linking: \"%s\"\n", */
/*                    code[i].nearest_func_name, log); */
          g_free (log);
        }

      /* Bilinear sampler */
      /* FIXME: Disabled ATM, switching to bilinear fixed func pipeline. */
/*       fragment_code = concatenate_sampler_code (fragment_header, */
/*                                                 code[i].sampler, */
/*                                                 fragment_main, */
/*                                                 code[i].bilinear_func_name); */
/*       fragment_shader = gl->create_shader_object (PGM_RENDER_GL_FRAGMENT_SHADER); */
/*       gl->shader_source (fragment_shader, */
/*                          get_shader_code_line_size */
/*                          ((const gchar**) fragment_code), */
/*                          (const gchar**) fragment_code, NULL); */
/*       free_shader_code (fragment_code); */
/*       gl->compile_shader (fragment_shader); */
/*       log = get_glsl_info_log (vertex_shader, gl); */
/*       if (log) */
/*         { */
/*           g_print ("\"%s\" fragment shader compilation: \"%s\"\n", */
/*                    code[i].bilinear_func_name, log); */
/*           g_free (log); */
/*         } */

/*       window->color_space_shader[i][1].program = gl->create_program_object (); */
/*       shader = window->color_space_shader[i][1]; */
/*       gl->attach_object (shader.program, vertex_shader); */
/*       gl->attach_object (shader.program, fragment_shader); */
/*       gl->link_program (shader.program); */
/*       log = get_glsl_info_log (shader.program, gl); */
/*       if (log) */
/*         { */
/*           g_print ("\"%s\" program linking: \"%s\"\n", */
/*                    code[i].bilinear_func_name, log); */
/*           g_free (log); */
/*         } */
    }

  return 0;
}

static void
opengl_extensions_proc_address_lookup (PgmRenderGlBackend *backend)
{
  PgmRenderGlProcAddressList *gl = backend->proc_address;

  /* Blend color */
  if (backend->feature_mask & PGM_RENDER_GL_FEATURE_BLEND_COLOR)
    {
      if (backend->version >= 1.4f)
        gl->blend_color = (pgm_render_gl_blend_color)
          backend->get_proc_address (backend, "glBlendColor");
      else
        gl->blend_color = (pgm_render_gl_blend_color)
          backend->get_proc_address (backend, "glBlendColorEXT");

      if (!gl->blend_color)
        backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_BLEND_COLOR;
    }

  /* Multi-texturing */
  if (backend->feature_mask & PGM_RENDER_GL_FEATURE_MULTITEXTURE)
    {
      if (backend->version >= 1.3f)
        {
          gl->active_texture = (pgm_render_gl_active_texture)
            backend->get_proc_address (backend, "glActiveTexture");
          gl->client_active_texture =
            (pgm_render_gl_client_active_texture)
            backend->get_proc_address (backend, "glClientActiveTexture");
	}
      else
        {
          gl->active_texture = (pgm_render_gl_active_texture)
            backend->get_proc_address (backend, "glActiveTextureARB");
          gl->client_active_texture =
            (pgm_render_gl_client_active_texture)
            backend->get_proc_address (backend, "glClientActiveTextureARB");
	}

      if ((!gl->active_texture)
          || (!gl->client_active_texture))
        {
          backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_MULTITEXTURE;
          backend->max_texture_units = 0;
	}
      else
        {
          gl->get_integer_v (PGM_RENDER_GL_MAX_TEXTURE_UNITS,
                             &backend->max_texture_units);
          if (backend->max_texture_units >= 3)
            backend->feature_mask |=
              PGM_RENDER_GL_FEATURE_PER_COMPONENT_RENDERING;
        }
    }

  /* Multi draw arrays */
  if (backend->feature_mask & PGM_RENDER_GL_FEATURE_MULTI_DRAW_ARRAYS)
    {
      if (backend->version >= 1.4f)
        gl->multi_draw_arrays = (pgm_render_gl_multi_draw_arrays)
          backend->get_proc_address (backend, "glMultiDrawArrays");
      else
        gl->multi_draw_arrays = (pgm_render_gl_multi_draw_arrays)
          backend->get_proc_address (backend, "glMultiDrawArraysEXT");

      if (!gl->multi_draw_arrays)
        backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_MULTI_DRAW_ARRAYS;
    }

  /* Fragment program */
  if (backend->feature_mask & PGM_RENDER_GL_FEATURE_FRAGMENT_PROGRAM)
    {
      gl->gen_programs = (pgm_render_gl_gen_programs)
        backend->get_proc_address (backend, "glGenProgramsARB");
      gl->delete_programs = (pgm_render_gl_delete_programs)
        backend->get_proc_address (backend, "glDeleteProgramsARB");
      gl->program_string = (pgm_render_gl_program_string)
        backend->get_proc_address (backend, "glProgramStringARB");
      gl->bind_program = (pgm_render_gl_bind_program)
        backend->get_proc_address (backend, "glBindProgramARB");
      gl->program_local_param_4fv =
        (pgm_render_gl_program_local_param_4fv)
        backend->get_proc_address (backend, "glProgramLocalParameter4fvARB");
      gl->get_program_iv = (pgm_render_gl_get_program_iv)
        backend->get_proc_address (backend, "glGetProgramivARB");

      if ((!gl->gen_programs)
          || (!gl->delete_programs)
          || (!gl->program_string)
          || (!gl->bind_program)
          || (!gl->program_local_param_4fv))
        backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_FRAGMENT_PROGRAM;
    }

  /* OpenGL Shading Language */
  if ((backend->feature_mask & PGM_RENDER_GL_FEATURE_SHADING_LANGUAGE)
      && (backend->feature_mask & PGM_RENDER_GL_FEATURE_SHADER_OBJECTS)
      && (backend->feature_mask & PGM_RENDER_GL_FEATURE_FRAGMENT_SHADER)
      && (backend->feature_mask & PGM_RENDER_GL_FEATURE_VERTEX_SHADER))
    {
      /* FIXME: Support the OpenGL 2.0 entry points. See
       * <http://developer.3dlabs.com/documents/GLmanpages/glMigrationGuide.htm>
       */
/*       if (backend->version >= 2.0f) */
/*         { */
/*         } */
/*       else */
/*         { */
      gl->delete_object = (pgm_render_gl_delete_object)
        backend->get_proc_address (backend, "glDeleteObjectARB");
      gl->get_handle = (pgm_render_gl_get_handle)
        backend->get_proc_address (backend, "glGetHandleARB");
      gl->detach_object = (pgm_render_gl_detach_object)
        backend->get_proc_address (backend, "glDetachObjectARB");
      gl->create_shader_object = (pgm_render_gl_create_shader_object)
        backend->get_proc_address (backend, "glCreateShaderObjectARB");
      gl->shader_source = (pgm_render_gl_shader_source)
        backend->get_proc_address (backend, "glShaderSourceARB");
      gl->compile_shader = (pgm_render_gl_compile_shader)
        backend->get_proc_address (backend, "glCompileShaderARB");
      gl->create_program_object = (pgm_render_gl_create_program_object)
        backend->get_proc_address (backend, "glCreateProgramObjectARB");
      gl->attach_object = (pgm_render_gl_attach_object)
        backend->get_proc_address (backend, "glAttachObjectARB");
      gl->link_program = (pgm_render_gl_link_program)
        backend->get_proc_address (backend, "glLinkProgramARB");
      gl->use_program_object = (pgm_render_gl_use_program_object)
        backend->get_proc_address (backend, "glUseProgramObjectARB");
      gl->get_object_parameter_iv = (pgm_render_gl_get_object_parameter_iv)
        backend->get_proc_address (backend, "glGetObjectParameterivARB");
      gl->get_info_log = (pgm_render_gl_get_info_log)
        backend->get_proc_address (backend, "glGetInfoLogARB");
      gl->get_uniform_location = (pgm_render_gl_get_uniform_location)
        backend->get_proc_address (backend, "glGetUniformLocationARB");
      gl->uniform_matrix_4fv = (pgm_render_gl_uniform_matrix_4fv)
        backend->get_proc_address (backend, "glUniformMatrix4fvARB");
      gl->uniform_1f = (pgm_render_gl_uniform_1f)
        backend->get_proc_address (backend, "glUniform1fARB");
      gl->uniform_1i = (pgm_render_gl_uniform_1i)
        backend->get_proc_address (backend, "glUniform1iARB");
/*         } */

      if ((!gl->delete_object)
          || (!gl->get_handle)
          || (!gl->detach_object)
          || (!gl->create_shader_object)
          || (!gl->shader_source)
          || (!gl->compile_shader)
          || (!gl->create_program_object)
          || (!gl->attach_object)
          || (!gl->link_program)
          || (!gl->use_program_object)
          || (!gl->get_object_parameter_iv)
          || (!gl->get_info_log)
          || (!gl->get_uniform_location)
          || (!gl->uniform_matrix_4fv)
          || (!gl->uniform_1f)
          || (!gl->uniform_1i))
        {
          backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_SHADING_LANGUAGE;
          backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_SHADER_OBJECTS;
          backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_FRAGMENT_SHADER;
          backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_VERTEX_SHADER;
        }
      else
        backend->feature_mask |= PGM_RENDER_GL_FEATURE_GLSL;
    }

  /* Vertex & Pixel Buffer Objects */
  if ((backend->feature_mask & PGM_RENDER_GL_FEATURE_VERTEX_BUFFER_OBJECT) ||
      (backend->feature_mask & PGM_RENDER_GL_FEATURE_PIXEL_BUFFER_OBJECT))
    {
      if (backend->version >= 1.5f)
        {
          gl->gen_buffers = (pgm_render_gl_gen_buffers)
            backend->get_proc_address (backend, "glGenBuffers");
          gl->delete_buffers = (pgm_render_gl_delete_buffers)
            backend->get_proc_address (backend, "glDeleteBuffers");
          gl->bind_buffer = (pgm_render_gl_bind_buffer)
            backend->get_proc_address (backend, "glBindBuffer");
          gl->buffer_data = (pgm_render_gl_buffer_data)
            backend->get_proc_address (backend, "glBufferData");
          gl->buffer_sub_data = (pgm_render_gl_buffer_sub_data)
            backend->get_proc_address (backend, "glBufferSubData");
          gl->get_buffer_sub_data = (pgm_render_gl_get_buffer_sub_data)
            backend->get_proc_address (backend, "glGetBufferSubData");
          gl->map_buffer = (pgm_render_gl_map_buffer)
            backend->get_proc_address (backend, "glMapBuffer");
          gl->unmap_buffer = (pgm_render_gl_unmap_buffer)
            backend->get_proc_address (backend, "glUnmapBuffer");
	}
      else
        {
          gl->gen_buffers = (pgm_render_gl_gen_buffers)
            backend->get_proc_address (backend, "glGenBuffersARB");
          gl->delete_buffers = (pgm_render_gl_delete_buffers)
            backend->get_proc_address (backend, "glDeleteBuffersARB");
          gl->bind_buffer = (pgm_render_gl_bind_buffer)
            backend->get_proc_address (backend, "glBindBufferARB");
          gl->buffer_data = (pgm_render_gl_buffer_data)
            backend->get_proc_address (backend, "glBufferDataARB");
          gl->buffer_sub_data = (pgm_render_gl_buffer_sub_data)
            backend->get_proc_address (backend, "glBufferSubDataARB");
          gl->get_buffer_sub_data = (pgm_render_gl_get_buffer_sub_data)
            backend->get_proc_address (backend, "glGetBufferSubDataARB");
          gl->map_buffer = (pgm_render_gl_map_buffer)
            backend->get_proc_address (backend, "glMapBufferARB");
          gl->unmap_buffer = (pgm_render_gl_unmap_buffer)
            backend->get_proc_address (backend, "glUnmapBufferARB");
	}

      if ((!gl->gen_buffers)
          || (!gl->delete_buffers)
          || (!gl->bind_buffer)
          || (!gl->buffer_data)
          || (!gl->buffer_sub_data)
          || (!gl->get_buffer_sub_data)
          || (!gl->map_buffer)
          || (!gl->unmap_buffer))
        {
          backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_VERTEX_BUFFER_OBJECT;
          backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_PIXEL_BUFFER_OBJECT;
        }
    }

  /* FrameBuffer Object */
  if (backend->feature_mask & PGM_RENDER_GL_FEATURE_FRAMEBUFFER_OBJECT)
    {
      gl->gen_framebuffers = (pgm_render_gl_gen_framebuffers)
        backend->get_proc_address (backend, "glGenFramebuffersEXT");
      gl->delete_framebuffers = (pgm_render_gl_delete_framebuffers)
        backend->get_proc_address (backend, "glDeleteFramebuffersEXT");
      gl->bind_framebuffer = (pgm_render_gl_bind_framebuffer)
        backend->get_proc_address (backend, "glBindFramebufferEXT");
      gl->framebuffer_renderbuffer =
        (pgm_render_gl_framebuffer_renderbuffer)
        backend->get_proc_address (backend, "glFramebufferRenderbufferEXT");
      gl->framebuffer_texture_2d =
        (pgm_render_gl_framebuffer_texture_2d)
        backend->get_proc_address (backend, "glFramebufferTexture2DEXT");
      gl->check_framebuffer_status =
        (pgm_render_gl_check_framebuffer_status)
        backend->get_proc_address (backend, "glCheckFramebufferStatusEXT");
      gl->gen_renderbuffers = (pgm_render_gl_gen_renderbuffers)
        backend->get_proc_address (backend, "glGenRenderbuffersEXT");
      gl->delete_renderbuffers = (pgm_render_gl_delete_renderbuffers)
        backend->get_proc_address (backend, "glDeleteRenderbuffersEXT");
      gl->bind_renderbuffer = (pgm_render_gl_bind_renderbuffer)
        backend->get_proc_address (backend, "glBindRenderbufferEXT");
      gl->renderbuffer_storage = (pgm_render_gl_renderbuffer_storage)
        backend->get_proc_address (backend, "glRenderbufferStorageEXT");
      gl->get_renderbuffer_parameter_iv =
        (pgm_render_gl_get_renderbuffer_parameter_iv)
        backend->get_proc_address (backend, "glGetRenderbufferParameterivEXT");

      if ((!gl->gen_framebuffers)
          || (!gl->delete_framebuffers)
          || (!gl->bind_framebuffer)
          || (!gl->framebuffer_renderbuffer)
          || (!gl->framebuffer_texture_2d)
          || (!gl->check_framebuffer_status)
          || (!gl->gen_renderbuffers)
          || (!gl->delete_renderbuffers)
          || (!gl->bind_renderbuffer)
          || (!gl->renderbuffer_storage)
          || (!gl->get_renderbuffer_parameter_iv))
        backend->feature_mask &= ~PGM_RENDER_GL_FEATURE_FRAMEBUFFER_OBJECT;
    }

  /* Maximum 2D and Rect texture size */
  gl->get_integer_v (PGM_RENDER_GL_MAX_TEXTURE_SIZE,
                     &backend->max_texture_2d_size);
  if (backend->feature_mask & PGM_RENDER_GL_FEATURE_TEXTURE_RECTANGLE)
    {
      gl->get_integer_v (PGM_RENDER_GL_MAX_RECTANGLE_TEXTURE_SIZE,
                         &backend->max_texture_rect_size);
    }
  else
    backend->max_texture_rect_size = 0;
  /* backend->max_texture_2d_size = 64; */

  /* Smooth line range */
  gl->get_float_v (PGM_RENDER_GL_SMOOTH_LINE_WIDTH_RANGE,
                   backend->smooth_line_range);

  /* Pixel stores default values */
  gl->get_integer_v (PGM_RENDER_GL_UNPACK_ROW_LENGTH, &backend->row_length);
  gl->get_integer_v (PGM_RENDER_GL_UNPACK_SKIP_ROWS, &backend->skip_rows);
  gl->get_integer_v (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, &backend->skip_pixels);
}

static gboolean
opengl_backend_init (PgmRenderGlBackend **backend)
{
  PgmRenderGlProcAddressList *gl;
  gint i = 0;

  /* Init OpenGL proc addresses */
  (*backend)->proc_address = &opengl_proc_address_list;
  gl = &opengl_proc_address_list;

  /* Retrieve OpenGL version */
  (*backend)->version_string =
    (const gchar *) gl->get_string (PGM_RENDER_GL_VERSION);
  if (!(*backend)->version_string)
    {
      g_warning ("Cannot retrieve OpenGL version.");
      return FALSE;
    }

  /* Check OpenGL version */
  (*backend)->version =
    opengl_version_string_to_float ((*backend)->version_string);
  if ((*backend)->version < 1.2f)
    {
      g_warning ("OpenGL version: %.1f (1.2 minimum required).",
                 (*backend)->version);
      return FALSE;
    }

  /* Retrieve renderer informations */
  (*backend)->vendor = (const gchar *) gl->get_string (PGM_RENDER_GL_VENDOR);
  (*backend)->renderer = (const gchar *) gl->get_string (PGM_RENDER_GL_RENDERER);
  (*backend)->extensions =
    (const gchar *) gl->get_string (PGM_RENDER_GL_EXTENSIONS);

  /* Retrieve the extensions and store the features */
  while (opengl_extensions_map[i].name)
    {
      if (check_opengl_extension ((*backend)->extensions,
                                  opengl_extensions_map[i].name))
        (*backend)->feature_mask |= opengl_extensions_map[i].mask;
      i++;
    }
 opengl_extensions_proc_address_lookup (*backend);

  return TRUE;
}

static void
opengl_backend_deinit (PgmRenderGlBackend **backend)
{
  g_free (*backend);
}

/* Define the OpenGL texture wrapping and minification/magnification filter
 * parameters to the actual binded texture. */
static inline void
set_texture_parameters (PgmRenderGLenum wrapping, PgmRenderGLenum filter,
                        PgmRenderGlProcAddressList *gl)
{
  gl->tex_parameter_i (PGM_RENDER_GL_TEXTURE_2D,
                       PGM_RENDER_GL_TEXTURE_WRAP_S, wrapping);
  gl->tex_parameter_i (PGM_RENDER_GL_TEXTURE_2D,
                       PGM_RENDER_GL_TEXTURE_WRAP_T, wrapping);
  gl->tex_parameter_i (PGM_RENDER_GL_TEXTURE_2D,
                       PGM_RENDER_GL_TEXTURE_MIN_FILTER, filter);
  gl->tex_parameter_i (PGM_RENDER_GL_TEXTURE_2D,
                       PGM_RENDER_GL_TEXTURE_MAG_FILTER, filter);
}

/* Restore the OpenGL pixel storing modes to the implementation default
 * values. */
static inline void
restore_pixel_store_modes (PgmRenderGlBackend *backend)
{
  PgmRenderGlProcAddressList *gl = backend->proc_address;

  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH,
                     backend->row_length);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS,
                     backend->skip_pixels);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS,
                     backend->skip_rows);
}

/* Classic RGB (24 bits) OpenGL texture creation. */
static void
create_rgb_texture (gint width, gint height, guint *tex_id,
                    PgmRenderGLenum filter, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_RGBA,
                    width, height, 0, PGM_RENDER_GL_RGB,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);
}

/* Classic BGR (24 bits) OpenGL texture creation. */
static void
create_bgr_texture (gint width, gint height, guint *tex_id,
                    PgmRenderGLenum filter, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_RGBA,
                    width, height, 0, PGM_RENDER_GL_BGR,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);
}

/* Classic RGBA (32 bits) OpenGL texture creation. */
static void
create_rgba_texture (gint width, gint height, guint *tex_id,
                     PgmRenderGLenum filter, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_RGBA,
                    width, height, 0, PGM_RENDER_GL_RGBA,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);
}

/* Classic BGRA (32 bits) OpenGL texture creation. */
static void
create_bgra_texture (gint width, gint height, guint *tex_id,
                     PgmRenderGLenum filter, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_RGBA,
                    width, height, 0, PGM_RENDER_GL_BGRA,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);
}

/* RGBA (32 bits) OpenGL texture creation to store packed 16 formats such as
 * UYVY and YUYV fourcc. Since we store 2 pixels in the 32 bits format, we
 * divide width by 2. */
static void
create_packed_16_texture (gint width, gint height, guint *tex_id,
                          PgmRenderGLenum filter,
                          PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_RGBA,
                    width * 0.5f, height, 0, PGM_RENDER_GL_RGBA,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);
}

/* 3 LUMINANCE (8 bits) textures creation to store respectively the Y, U and V
 * planes of planar formats such as I420 and YV12. */
static void
create_planar_12_per_plane_texture (gint width, gint height, guint *tex_id,
                                    PgmRenderGLenum filter,
                                    PgmRenderGlProcAddressList *gl)
{
  /* Y plane */
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, tex_id[0]);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_LUMINANCE,
                    width, height, 0, PGM_RENDER_GL_LUMINANCE,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);

  /* U plane */
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, tex_id[1]);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_LUMINANCE,
                    width * 0.5f, height * 0.5f, 0, PGM_RENDER_GL_LUMINANCE,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);

  /* V plane */
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, tex_id[2]);
  gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_LUMINANCE,
                    width * 0.5f, height * 0.5f, 0, PGM_RENDER_GL_LUMINANCE,
                    PGM_RENDER_GL_UNSIGNED_BYTE, NULL);
  set_texture_parameters (PGM_RENDER_GL_CLAMP, filter, gl);
}

/* Classic RGB (24 bits) OpenGL texture data upload. */
static void
upload_rgb_texture (gint width, gint height, gint stride, gint row_lenght,
                    gint skip_rows, gint skip_pixels, guint *tex_id,
                    void *data, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels);
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0, width, height,
                        PGM_RENDER_GL_RGB, PGM_RENDER_GL_UNSIGNED_BYTE, data);
}

/* Classic BGR (24 bits) OpenGL texture data upload. */
static void
upload_bgr_texture (gint width, gint height, gint stride, gint row_lenght,
                    gint skip_rows, gint skip_pixels, guint *tex_id,
                    void *data, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels);
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0, width, height,
                        PGM_RENDER_GL_BGR, PGM_RENDER_GL_UNSIGNED_BYTE, data);
}

/* Classic RGBA (32 bits) OpenGL texture data upload. */
static void
upload_rgba_texture (gint width, gint height, gint stride, gint row_lenght,
                     gint skip_rows, gint skip_pixels, guint *tex_id,
                     void *data, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels);
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0, width, height,
                        PGM_RENDER_GL_RGBA, PGM_RENDER_GL_UNSIGNED_BYTE, data);
}

/* Classic BGRA (32 bits) OpenGL texture data upload. */
static void
upload_bgra_texture (gint width, gint height, gint stride, gint row_lenght,
                     gint skip_rows, gint skip_pixels, guint *tex_id,
                     void *data, PgmRenderGlProcAddressList *gl)
{
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows);
  gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels);
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0, width, height,
                        PGM_RENDER_GL_BGRA, PGM_RENDER_GL_UNSIGNED_BYTE, data);
}

/* */
/* FIXME: Doesn't work with texture tiling */
static void
upload_packed_16_texture (gint width, gint height, gint stride, gint row_lenght,
                          gint skip_rows, gint skip_pixels, guint *tex_id,
                          void *data, PgmRenderGlProcAddressList *gl)
{
/*   g_print ("upload_packed_16_texture (w=%d h=%d rl=%d sr=%d sp=%d)\n", */
/*            width, height, row_lenght, skip_rows, skip_pixels); */

  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, *tex_id);
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels); */
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0, width * 0.5f, height,
                        PGM_RENDER_GL_RGBA, PGM_RENDER_GL_UNSIGNED_BYTE, data);
}

/* */
/* FIXME: Doesn't work with texture tiling */
static void
upload_planar_12_per_plane_texture (gint width, gint height, gint stride,
                                    gint row_lenght, gint skip_rows,
                                    gint skip_pixels, guint *tex_id,
                                    void *data, PgmRenderGlProcAddressList *gl)
{
  gint offset;

/*   g_print ("upload_planar_12_per_plane_texture (w=%d h=%d rl=%d sr=%d sp=%d)\n", */
/*            width, height, row_lenght, skip_rows, skip_pixels); */

  /* Y plane */
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, tex_id[0]);
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels); */
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0,
                        width, height, PGM_RENDER_GL_LUMINANCE,
                        PGM_RENDER_GL_UNSIGNED_BYTE, data);

  /* U plane */
  offset = width * height;
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, tex_id[1]);
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght * 0.5f); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows * 0.5f); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels * 0.5f); */
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0,
                        width * 0.5f, height * 0.5f, PGM_RENDER_GL_LUMINANCE,
                        PGM_RENDER_GL_UNSIGNED_BYTE, (void*) ((unsigned char*) data + offset));

  /* V plane */
  offset += width * height * 0.25f;
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, tex_id[2]);
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_ROW_LENGTH, row_lenght * 0.5f); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_ROWS, skip_rows * 0.5f); */
/*   gl->pixel_store_i (PGM_RENDER_GL_UNPACK_SKIP_PIXELS, skip_pixels * 0.5f); */
  gl->tex_sub_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, 0, 0,
                        width * 0.5f, height * 0.5f, PGM_RENDER_GL_LUMINANCE,
                        PGM_RENDER_GL_UNSIGNED_BYTE, (void*) ((unsigned char*) data + offset));
}

static void
pgm_render_gl1_window_class_init (PgmRenderGl1WindowClass *klass)
{
  GObjectClass *gobject_class;
  PgmRenderBaseWindowClass *basewindow_class;

  gobject_class = (GObjectClass*) klass;
  basewindow_class = (PgmRenderBaseWindowClass*) klass;

  basewindow_class->update = pgm_render_gl1_window_update;
  basewindow_class->set_color = pgm_render_gl1_window_set_color;
  basewindow_class->get_color = pgm_render_gl1_window_get_color;
  basewindow_class->set_title = pgm_render_gl1_window_set_title;
  basewindow_class->set_size = pgm_render_gl1_window_set_size;
  basewindow_class->set_canvas_size = pgm_render_gl1_window_set_canvas_size;
  basewindow_class->set_fullscreen = pgm_render_gl1_window_set_fullscreen;
  basewindow_class->get_fullscreen = pgm_render_gl1_window_get_fullscreen;
  basewindow_class->get_events = pgm_render_gl1_window_get_events;
  basewindow_class->start_loop = pgm_render_gl1_window_start_loop;
  basewindow_class->quit_loop = pgm_render_gl1_window_quit_loop;
  basewindow_class->add_drawable = pgm_render_gl1_window_add_drawable;
  basewindow_class->del_drawable = pgm_render_gl1_window_del_drawable;
  basewindow_class->reorder_drawable = pgm_render_gl1_window_reorder_drawable;
  basewindow_class->regenerate_drawables =
    pgm_render_gl1_window_regenerate_drawables;
  basewindow_class->set_scale = pgm_render_gl1_window_set_scale;
  basewindow_class->set_scale_mode = pgm_render_gl1_window_set_scale_mode;
  basewindow_class->set_cursor = pgm_render_gl1_window_set_cursor;
  basewindow_class->get_screen_size = pgm_render_gl1_window_get_screen_size;
  basewindow_class->get_screen_size_mm =
    pgm_render_gl1_window_get_screen_size_mm;
  basewindow_class->get_screen_resolution =
    pgm_render_gl1_window_get_screen_resolution;

  klass->surface_class = NULL;
  klass->font_class = NULL;

  gobject_class->finalize = pgm_render_gl1_window_finalize;
}

static void
pgm_render_gl1_window_init (PgmRenderGl1Window *obj)
{
  obj->context_creation_mutex = g_mutex_new ();
  obj->context_creation_cond = g_cond_new ();
  obj->start_loop_mutex = g_mutex_new ();
  obj->start_loop_cond = g_cond_new ();
  obj->loop_started = FALSE;
  obj->context_created = FALSE;
  obj->window_title = g_strdup ("");
  obj->icon_title = g_strdup ("");
  obj->event = NULL;
  obj->fullscreen = FALSE;
  obj->new_size = FALSE;
  obj->new_canvas_size = FALSE;
  obj->new_refresh_color = FALSE;
  obj->new_cursor = FALSE;
  obj->color.r = obj->color.g = obj->color.b = 0.0f;
  obj->color.a = 1.0f;
  obj->source_id = 0;
  obj->source = NULL;
  obj->creation_filter = PGM_RENDER_GL_LINEAR;
  obj->create_texture[PGM_RENDER_MEDIA_RGB] = create_rgb_texture;
  obj->create_texture[PGM_RENDER_MEDIA_BGR] = create_bgr_texture;
  obj->create_texture[PGM_RENDER_MEDIA_RGBA] = create_rgba_texture;
  obj->create_texture[PGM_RENDER_MEDIA_BGRA] = create_bgra_texture;
  obj->create_texture[PGM_RENDER_MEDIA_I420] =
    create_planar_12_per_plane_texture;
  obj->create_texture[PGM_RENDER_MEDIA_YV12] =
    create_planar_12_per_plane_texture;
  obj->create_texture[PGM_RENDER_MEDIA_UYVY] = create_packed_16_texture;
  obj->create_texture[PGM_RENDER_MEDIA_YUYV] = create_packed_16_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_RGB] = upload_rgb_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_BGR] = upload_bgr_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_RGBA] = upload_rgba_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_BGRA] = upload_bgra_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_I420] =
    upload_planar_12_per_plane_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_YV12] =
    upload_planar_12_per_plane_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_UYVY] =
    upload_packed_16_texture;
  obj->upload_texture[PGM_RENDER_MEDIA_YUYV] =
    upload_packed_16_texture;
  obj->new_scale = FALSE;
  obj->new_title = FALSE;
  obj->fullscreen_request = FALSE;
  g_get_current_time (&obj->update_time);
  g_get_current_time (&obj->fps_print_time);
  obj->show_fps = FALSE;
  obj->surface = NULL;
  obj->font = NULL;
  g_window = obj;
}

static void
pgm_render_gl1_window_finalize (GObject *obj)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (obj);

  g_mutex_free (gl1_window->context_creation_mutex);
  g_cond_free (gl1_window->context_creation_cond);
  g_mutex_free (gl1_window->start_loop_mutex);
  g_cond_free (gl1_window->start_loop_cond);

/*   gl1_window->backend->destroy_text_lists (gl1_window->backend); */
  gl1_window->backend->destroy_context (gl1_window->backend);
  gl1_window->backend->destroy_window (gl1_window->backend);
  opengl_backend_deinit (&gl1_window->backend);

  G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static void
reshape (PgmRenderGlProcAddressList *gl, gint canvas_width, gint canvas_height,
         gint window_width, gint window_height)
{
  gfloat fov = 0.15f;
  gfloat near_plane = 500.0f;
  gfloat far_plane = 5000.0f;
  gfloat tan_fov = tan (fov / (360.0f / G_PI));
  gfloat distance = 1.0f / tan_fov;
  gfloat n = (distance - near_plane) * tan_fov;

  gl->viewport (0, window_height-canvas_height, canvas_width, canvas_height);
  gl->matrix_mode (PGM_RENDER_GL_PROJECTION);
  gl->load_identity ();
  gl->frustum (-n, n, -n, n, distance-near_plane, distance+far_plane);
  gl->scale_f (2.0f/canvas_width, -2.0f/canvas_height, 1.0f);
  gl->translate_f (-canvas_width/2.0f, -canvas_height/2.0f, -distance);
  gl->matrix_mode (PGM_RENDER_GL_MODELVIEW);
}

static void
init_opengl_states (PgmRenderGlProcAddressList *gl)
{
  gl->enable (PGM_RENDER_GL_DEPTH_TEST);
  gl->shade_model (PGM_RENDER_GL_SMOOTH);
  gl->enable (PGM_RENDER_GL_BLEND);
  gl->blend_func (PGM_RENDER_GL_SRC_ALPHA, PGM_RENDER_GL_ONE_MINUS_SRC_ALPHA);
  gl->clear (PGM_RENDER_GL_COLOR_BUFFER_BIT | PGM_RENDER_GL_DEPTH_BUFFER_BIT);
  gl->clear_color (0.0f ,0.0f, 0.0f, 1.0f);
}

static void
dump_opengl_features (PgmRenderGlBackend *backend)
{
  gulong mask = backend->feature_mask;
  gboolean glsl_shader = mask & PGM_RENDER_GL_FEATURE_FRAGMENT_PROGRAM;
  gboolean per_yuv_units = mask & PGM_RENDER_GL_FEATURE_PER_COMPONENT_RENDERING;

  g_print (" --- Pigment OpenGL status ------------------------\n"
           " Version:  %s\n"
           " Renderer: %s\n"
           " Vendor:   %s\n"
           " GLSL shaders: .............. %s (%s)\n"
           " Per YUV component rendering: %s (%s)\n"
           " Texture rectangle: ......... %s (not implemented)\n"
           " Fragment program: .......... %s (not implemented)\n"
           " Pixel buffer object: ....... %s (not implemented)\n"
           " --------------------------------------------------\n",
           backend->version_string, backend->renderer, backend->vendor,
           glsl_shader ? "Yes" : "No ",
           glsl_shader && per_yuv_units ? "used" : "not used",
           per_yuv_units ? "Yes" : "No ",
           glsl_shader && per_yuv_units ? "used" : "not used",
           (mask & PGM_RENDER_GL_FEATURE_TEXTURE_RECTANGLE) ? "Yes" : "No ",
           (mask & PGM_RENDER_GL_FEATURE_FRAGMENT_PROGRAM) ? "Yes" : "No ",
           (mask & PGM_RENDER_GL_FEATURE_PIXEL_BUFFER_OBJECT) ? "Yes" : "No ");
}

static void
dump_gpu_support (PgmRenderGlBackend *backend)
{
  gulong mask = backend->feature_mask;
  gint renderer_length = strlen (backend->renderer);
  gint i;

  g_print ("\n--- Pigment GPU Support Dump ---\n%s\n", backend->renderer);
  for (i = 0; i < renderer_length; i++)
    g_print ("-");
  g_print ("\n\n**Version:**\n%s\n\n**Vendor:**\n%s\n\n**Features:**\n\n::\n\n",
           backend->version_string, backend->vendor);
  g_print ("  maximum_texture_2d_size:  %d\n  texture_rectangle: ...... %s\n"
           "  texture_non_power_of_two: %s\n  texture_border_clamp: ... %s\n"
           "  multisample: ............ %s\n  multisample_filter_hint:  %s\n"
           "  multitexture: ........... %s (%d units)\n"
           "  texture_env_combine: .... %s\n  texture_env_dot3: ....... %s\n"
           "  fragment_program: ....... %s\n  vertex_buffer_object: ... %s\n"
           "  pixel_buffer_object: .... %s\n  blend_color: ............ %s\n"
           "  packed_pixels: .......... %s\n  multi_draw_arrays: ...... %s\n"
           "  framebuffer_object: ..... %s\n  copy_sub_buffer: ........ %s\n"
           "  shading_language: ....... %s\n  shader_objects: ......... %s\n"
           "  fragment_shader: ........ %s\n  vertex_shader: .......... %s\n",
           backend->max_texture_2d_size,
           mask & PGM_RENDER_GL_FEATURE_TEXTURE_RECTANGLE ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_TEXTURE_NON_POWER_OF_TWO ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_TEXTURE_BORDER_CLAMP ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_MULTISAMPLE ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_MULTISAMPLE_FILTER_HINT ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_MULTITEXTURE ? "1" : "0",
           backend->max_texture_units,
           mask & PGM_RENDER_GL_FEATURE_TEXTURE_ENV_COMBINE ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_TEXTURE_ENV_DOT3 ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_FRAGMENT_PROGRAM ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_VERTEX_BUFFER_OBJECT ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_PIXEL_BUFFER_OBJECT ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_BLEND_COLOR ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_PACKED_PIXELS ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_MULTI_DRAW_ARRAYS ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_FRAMEBUFFER_OBJECT ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_COPY_SUB_BUFFER ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_SHADING_LANGUAGE ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_SHADER_OBJECTS ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_FRAGMENT_SHADER ? "1" : "0",
           mask & PGM_RENDER_GL_FEATURE_VERTEX_SHADER ? "1" : "0");
  g_print ("--------------------------------\n\n");
}

static void
_get_gst_image_from_surface (PgmRenderGl1Surface *gl1_surface,
                             GstBuffer **buffer)
{
  PGM_RENDER_DRAWABLE_LOCK (gl1_surface);
  *buffer = gl1_surface->media_gst_buffer;
  if (*buffer)
    gst_buffer_ref (*buffer);
  PGM_RENDER_DRAWABLE_UNLOCK (gl1_surface);
}

static gint
_refresh_gst_image (PgmRenderGl1Window *window, PgmRenderGl1Surface *surface)
{
  PgmRenderGlBackend *backend = window->backend;
  PgmRenderGlProcAddressList *gl = backend->proc_address;
  GstBuffer *buffer = NULL;
  gint tex_width, tex_height, sub_tex_width, sub_tex_height;
  guint *tex_id;
  gint i, j;
  PgmRenderGLenum filter = PGM_RENDER_GL_LINEAR;

  if (!PGM_RENDER_IS_GL1_SURFACE (surface))
    return 1;

  _get_gst_image_from_surface (surface, &buffer);

  if (!buffer)
    return 2;

  /* The texture objects have not already been created. */
  if (!surface->media_gst_loaded)
    {
      tex_width = MIN (surface->media_pot_width,
                       surface->max_texture_size);
      tex_height = MIN (surface->media_pot_height,
                        surface->max_texture_size);
      gl->gen_textures (surface->media_texture->size,
                        surface->media_texture->identifiants);
      if (surface->media_format == PGM_RENDER_MEDIA_YUYV
          || surface->media_format == PGM_RENDER_MEDIA_UYVY)
        filter = PGM_RENDER_GL_NEAREST;
      for (i = 0; i < surface->media_nb_width_tiles; i++)
        {
          for (j = 0; j < surface->media_nb_height_tiles; j++)
            {
              tex_id = &surface->media_texture->identifiants
                [j + surface->media_nb_height_tiles * i];
              window->create_texture[surface->media_format]
                (tex_width, tex_height, tex_id, filter, gl);
            }
        }
      surface->media_gst_loaded = TRUE;
    }

  /* Upload the video frame(s) */
  tex_width = MIN (surface->media_width, surface->max_texture_size);
  tex_height = MIN (surface->media_height, surface->max_texture_size);
  sub_tex_width = tex_width;
  sub_tex_height = tex_height;
  for (i = 0; i < surface->media_nb_width_tiles; i++)
    {
      if (i == surface->media_nb_width_tiles - 1)
        sub_tex_width = surface->media_width - i * tex_width;
      for (j = 0; j < surface->media_nb_height_tiles; j++)
        {
          if (j == surface->media_nb_height_tiles - 1)
            sub_tex_height = surface->media_height - j * tex_height;
          tex_id = &surface->media_texture->identifiants
            [j + surface->media_nb_height_tiles * i];
          window->upload_texture[surface->media_format]
            (sub_tex_width, sub_tex_height, surface->media_stride,
             surface->media_width, j * tex_width, i * tex_width, tex_id,
             (void*) GST_BUFFER_DATA (buffer), gl);
          sub_tex_height = tex_height;
        }
      sub_tex_width = tex_width;
    }
      
  restore_pixel_store_modes (backend);

  gst_buffer_unref (buffer);

  return 0;
}

static void
_set_interpolation_mode (PgmRenderGl1SurfaceClass *klass,
                         PgmRenderGl1Window *window)
{
  PgmRenderGlProcAddressList *gl = window->backend->proc_address;
  GList *element = klass->surface_with_new_media_interp;
  PgmRenderGl1Surface *surface;
  PgmRenderGLenum filter;
  gint i;

  while (element)
    {
      surface = PGM_RENDER_GL1_SURFACE (element->data);
      if (surface->media_interp == PGM_RENDER_MEDIA_NEAREST)
        filter = PGM_RENDER_GL_NEAREST;
      else
        filter = PGM_RENDER_GL_LINEAR;

      for (i = 0; i < surface->media_texture->size; i++)
        {
          gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D,
                            surface->media_texture->identifiants[i]);
          gl->tex_parameter_i (PGM_RENDER_GL_TEXTURE_2D,
                               PGM_RENDER_GL_TEXTURE_MIN_FILTER, filter);
          gl->tex_parameter_i (PGM_RENDER_GL_TEXTURE_2D,
                               PGM_RENDER_GL_TEXTURE_MAG_FILTER, filter);
        }

      element = g_list_next (element);
    }

  g_list_free (klass->surface_with_new_media_interp);
  klass->surface_with_new_media_interp = NULL;
}

static void
_free_static_textures (PgmRenderGl1SurfaceClass *klass,
                       PgmRenderGlProcAddressList *gl)
{
  GList *element = klass->texture_trash;
  PgmRenderTextureObject *texture;

  while (element)
    {
      texture = (PgmRenderTextureObject*) element->data;

      /* Free video memory */
      gl->delete_textures (texture->size, texture->identifiants);
      /* Free texture id */
      g_free (texture->identifiants);
      texture->identifiants = NULL;
      /* Free texture object */
      g_free (texture);
      texture = NULL;

      element = g_list_next (element);
    }

  /* Free the trash list */
  g_list_free (klass->texture_trash);
  klass->texture_trash = NULL;
}

static void
_sync_static_textures (PgmRenderGl1SurfaceClass *klass,
                       PgmRenderGl1Window *window)
{
  PgmRenderGlBackend *backend = window->backend;
  PgmRenderGlProcAddressList *gl = backend->proc_address;
  GList *element = g_list_reverse (klass->surface_with_media_data);
  PgmRenderGl1Surface *surface;
  gint tex_width, tex_height, sub_tex_width, sub_tex_height;
  guint *tex_id;
  gint i, j;

  /* Parse all the surface containing new media data */
  while (element)
    {
      surface = PGM_RENDER_GL1_SURFACE (element->data);

      /* Are there new textures to upload ? */
      if (surface->media_data != NULL)
        {
          tex_width = MIN (surface->media_pot_width,
                           surface->max_texture_size);
          tex_height = MIN (surface->media_pot_height,
                            surface->max_texture_size);
          sub_tex_width = tex_width;
          sub_tex_height = tex_height;
          gl->gen_textures (surface->media_texture->size,
                            surface->media_texture->identifiants);
          /* Send texture tiles to video memory */
          for (i = 0; i < surface->media_nb_width_tiles; i++)
            {
              if (i == surface->media_nb_width_tiles - 1)
                sub_tex_width = surface->media_width - i * tex_width;
              for (j = 0; j < surface->media_nb_height_tiles; j++)
                {
                  if (j == surface->media_nb_height_tiles - 1)
                    sub_tex_height = surface->media_height - j * tex_height;
                  tex_id = &surface->media_texture->identifiants
                    [j + surface->media_nb_height_tiles * i];
                  window->create_texture[surface->media_format]
                    (tex_width, tex_height, tex_id, window->creation_filter,
                     gl);
                  window->upload_texture[surface->media_format]
                    (sub_tex_width, sub_tex_height, surface->media_stride,
                     surface->media_width, j * tex_width, i * tex_width, tex_id,
                     (void*) surface->media_data, gl);
                  sub_tex_height = tex_height;
                }
              sub_tex_width = tex_width;
            }

          restore_pixel_store_modes (backend);
          g_free (surface->media_data);
          surface->media_data = NULL;
          surface->media_loaded = TRUE;
        }
      element = g_list_next (element);
    }

  g_list_free (klass->surface_with_media_data);
  klass->surface_with_media_data = NULL;
}

static void
_sync_font_textures (PgmRenderGl1FontClass *klass,
                     PgmRenderGlBackend *backend)
{
  PgmRenderGlProcAddressList *gl = backend->proc_address;
  GList *element = g_list_reverse (klass->font_with_media_data);
  PgmRenderGl1Font *font;

  while (element)
    {
      font = PGM_RENDER_GL1_FONT (element->data);

      gl->delete_textures (1, &font->text_id);
      gl->gen_textures (1, &font->text_id);
      gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, font->text_id);
      set_texture_parameters (PGM_RENDER_GL_CLAMP_TO_BORDER,
                              PGM_RENDER_GL_LINEAR, gl);
      gl->tex_image_2d (PGM_RENDER_GL_TEXTURE_2D, 0, PGM_RENDER_GL_RGBA,
                        font->text_tex_width, font->text_tex_height, 0,
                        PGM_RENDER_GL_BGRA, PGM_RENDER_GL_UNSIGNED_BYTE,
                        (void*) font->text_pixels);

      element = g_list_next (element);
    }

  g_list_free (klass->font_with_media_data);
  klass->font_with_media_data = NULL;
}

/* static void */
/* _sync_dynamic_textures (PgmRenderGl1Surface *surface) */
/* { */
/*   return; */
/* } */

static inline void
raster_fps_string (PgmRenderGl1Window *window, gfloat refresh_rate)
{
  GTimeVal current_time;
  gchar fps_string[16];
  gfloat elapsed_time;
  static gint fps;

  /* Retrieve current time and compute the time elapsed since the refresh */
  g_get_current_time (&current_time);
  elapsed_time = (current_time.tv_sec - window->fps_print_time.tv_sec)
    + (current_time.tv_usec - window->fps_print_time.tv_usec) * 0.000001f;

  /* Is the refresh_rate elapsed? */
  if (elapsed_time > refresh_rate)
    {
      /* Compute the time elapsed since the last update */
      elapsed_time = (current_time.tv_sec - window->update_time.tv_sec)
        + (current_time.tv_usec - window->update_time.tv_usec) * 0.000001f;
      fps = 1.0f / elapsed_time;

      /* Set the new FPS print time */
      window->fps_print_time.tv_sec = current_time.tv_sec;
      window->fps_print_time.tv_usec = current_time.tv_usec;
    }

  /* Build and rasterize the FPS string */
  g_snprintf (fps_string, 16, "%d FPS", fps);
  window->backend->raster_text (window->backend, fps_string, 6, 15,
                                0.0f, 0.0f, 0.0f);
  window->backend->raster_text (window->backend, fps_string, 5, 14,
                                1.0f, 1.0f, 1.0f);

  /* Set the new update time */
  window->update_time.tv_sec = current_time.tv_sec;
  window->update_time.tv_usec = current_time.tv_usec;
}

static gboolean
render (gpointer data)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (data);
  PgmRenderBaseWindow *window = PGM_RENDER_BASE_WINDOW (gl1_window);
  PgmRenderGl1WindowClass *gl1_window_class =
    PGM_RENDER_GL1_WINDOW_GET_CLASS (gl1_window);
  PgmRenderGl1SurfaceClass *gl1_surface_class = gl1_window_class->surface_class;
  PgmRenderGl1FontClass *gl1_font_class = gl1_window_class->font_class;
  PgmRenderGlProcAddressList *gl = gl1_window->backend->proc_address;
  GList *element;

  PGM_RENDER_BASE_WINDOW_LOCK (window);
  /* FIXME: We MUST use a queue to handle all these cases. */
  if (gl1_window->new_scale)
    {
      reshape (gl, window->canvas_width, window->canvas_height, window->width,
               window->height);
      gl1_window->new_scale = FALSE;
    }
  if (gl1_window->new_size)
    {
      if (window->scale_mode == PGM_RENDER_WINDOW_ADJUST)
        {
          _set_scale (gl1_window, (gfloat) window->width /
                      (gfloat) window->canvas_width,
                      (gfloat) window->height / (gfloat) window->canvas_height);
          reshape (gl, window->width, window->height, window->width,
                   window->height);
        }
      else if (window->scale_mode == PGM_RENDER_WINDOW_SCREEN_ADJUST)
        {
          _set_scale (gl1_window, (gfloat) window->width /
                      (gfloat) window->canvas_width,
                      (gfloat) window->height / (gfloat) window->canvas_height);
          reshape (gl, window->width, window->height, window->width,
                   window->height);
        }
      else
        reshape (gl, window->canvas_width, window->canvas_height, window->width,
                 window->height);
      gl1_window->backend->set_size (gl1_window->backend, window->width,
                                     window->height);
      gl1_window->new_size = FALSE;
    }
  if (gl1_window->new_canvas_size)
    {
      if (window->scale_mode == PGM_RENDER_WINDOW_ADJUST)
        _set_scale (gl1_window, (gfloat) window->width
                    / (gfloat) window->canvas_width,
                    (gfloat) window->height / (gfloat) window->canvas_height);
      gl1_window->new_canvas_size = FALSE;
    }
  if (gl1_window->new_refresh_color)
    {
      gl->clear_color (gl1_window->color.r, gl1_window->color.g,
                       gl1_window->color.b, gl1_window->color.a);
      gl1_window->new_refresh_color = FALSE;
    }
  if (gl1_window->new_title)
    {
      gl1_window->backend->set_title (gl1_window->backend,
                                      gl1_window->window_title,
                                      gl1_window->icon_title);
      gl1_window->new_title = FALSE;
    }
  if (gl1_window->fullscreen_request)
    {
      gl1_window->backend->set_fullscreen (gl1_window->backend,
                                           gl1_window->fullscreen,
                                           window->width,
                                           window->height);
      reshape (gl1_window->backend->proc_address, window->canvas_width,
               window->canvas_height, window->width, window->height);
      gl1_window->fullscreen_request = FALSE;
    }
  if (gl1_window->new_cursor == TRUE)
    {
      gl1_window->backend->set_cursor (gl1_window->backend, window->cursor);
      gl1_window->new_cursor = FALSE;
    }
  PGM_RENDER_BASE_WINDOW_UNLOCK (window);

  /* FIXME: We MUST wait on a condition */
  if (window->dirty)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      /* Texture management */
      if (gl1_surface_class)
        {
          _free_static_textures (gl1_surface_class, gl);
          _sync_static_textures (gl1_surface_class, gl1_window);
          _set_interpolation_mode (gl1_surface_class, gl1_window);
        }
      if (gl1_font_class)
        {
          /* free_font_textures (gl1_font_class, gl); */
          _sync_font_textures (gl1_font_class, gl1_window->backend);
        }

      /* Draw the far elements */
      gl->disable (PGM_RENDER_GL_DEPTH_TEST);
      element = window->far_drawable;
      while (element)
	{
          if (PGM_RENDER_DRAWABLE (element->data)->show)
            {
              _refresh_gst_image (gl1_window, element->data);
              pgm_render_drawable_draw (element->data);
            }
	  element = g_list_next (element);
	}
      /* Draw the z-ordered elements */
      gl->enable (PGM_RENDER_GL_DEPTH_TEST);
      element = window->z_ordered_drawable;
      while (element)
	{
          if (PGM_RENDER_DRAWABLE (element->data)->show)
            {
              _refresh_gst_image (gl1_window, element->data);
              pgm_render_drawable_draw (element->data);
            }
	  element = g_list_next (element);
	}
      /* Draw the near elements */
      gl->disable (PGM_RENDER_GL_DEPTH_TEST);
      element = window->near_drawable;
      while (element)
	{
          if (PGM_RENDER_DRAWABLE (element->data)->show)
            {
              _refresh_gst_image (gl1_window, element->data);
              pgm_render_drawable_draw (element->data);
            }
	  element = g_list_next (element);
	}

      /* Show FPS */
      if (gl1_window->show_fps)
        raster_fps_string (gl1_window, 0.5f);

      /* Swap the buffers and clear */
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
      /* FIXME: Temporary disabling the VBlank sync */
      /* gl1_window->backend->wait_for_vblank (gl1_window->backend); */
      gl1_window->backend->swap_buffers (gl1_window->backend);
      gl->clear (PGM_RENDER_GL_COLOR_BUFFER_BIT
                 | PGM_RENDER_GL_DEPTH_BUFFER_BIT);
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->dirty = FALSE;
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
    }

  return TRUE;
}

static gpointer
render_loop (gpointer data)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (data);
  PgmRenderBaseWindow *base_window = PGM_RENDER_BASE_WINDOW (data);
  GSource *source;
  PgmRenderEventSource *event_source;
  const gchar *env_var;
  gint creation;

  /* Allocate memory for the OpenGL backend */
  gl1_window->backend = 
    (PgmRenderGlBackend*) g_malloc0 (sizeof (PgmRenderGlBackend));

  /* Initialize GL window management functions */
#if defined(G_OS_WIN32)
  wgl_init_context (gl1_window);
#elif defined(G_OS_UNIX)
  glx_init_context (gl1_window);
#endif
  gl1_window->backend->resolution_width = 4;
  gl1_window->backend->resolution_height = 3;
  gl1_window->backend->screen_width_mm = 4;
  gl1_window->backend->screen_height_mm = 3;

  /* Create window and context */
  creation = gl1_window->backend->create_window (gl1_window->backend);
  creation &= gl1_window->backend->set_pixel_format (gl1_window->backend);
  creation &= gl1_window->backend->create_context (gl1_window->backend);
  creation &= gl1_window->backend->make_current (gl1_window->backend);
  if (!creation)
    g_error ("Unable to create the window.");
  gl1_window->backend->get_display_data (gl1_window->backend);

  /* Set up the OpenGL backend */
  if (!opengl_backend_init (&gl1_window->backend))
    g_error ("Unable to init OpenGL backend.");

  /* Load rasterized text display lists */
/*   if (!gl1_window->backend->build_text_lists (gl1_window->backend)) */
/*     g_error ("Unable to create text lists."); */

  /* Dump used features on stdout */
  dump_opengl_features (gl1_window->backend);

  /* Dump the GPU support info in the Rest format to stdout, used by the
   * Pigment GPU Support page.*/
  env_var = g_getenv (DUMP_GPU_SUPPORT_ENV_NAME);
  if (env_var && env_var[0] == '1')
    dump_gpu_support (gl1_window->backend);

  /* Enable FPS showing is the "PGM_SHOW_FPS" is set to 1 */
  /* FIXME: It should be done through a command line option */
/*   env_var = g_getenv (SHOW_FPS_ENV_NAME); */
/*   if (env_var && env_var[0] == '1') */
/*     gl1_window->show_fps = TRUE; */

  /* Build the shaders if the GPU is programmable */
  /* FIXME: We only active shaders if the number of texture units >= 3.
   *        The good way is to support packed 12 fourcc with one texture
   *        unit. */
  if (gl1_window->backend->feature_mask & PGM_RENDER_GL_FEATURE_GLSL
      && gl1_window->backend->feature_mask
      & PGM_RENDER_GL_FEATURE_PER_COMPONENT_RENDERING)
    {
      build_color_space_shaders (gl1_window);
      /* We are doing the filtering in the shaders, so there's no need to
       * set the OpenGL texture min/mag filter parameter to bilinear. */
      /* gl1_window->creation_filter = PGM_RENDER_GL_NEAREST; */
    }

  /* Initialize the OpenGL states and set the projection matrix */
  init_opengl_states (gl1_window->backend->proc_address);
  reshape (gl1_window->backend->proc_address, base_window->canvas_width,
           base_window->canvas_height, base_window->width,
           base_window->height);

  /* Signal the window creation func the end of the context creation */
  g_mutex_lock (gl1_window->context_creation_mutex);
  gl1_window->context_created = TRUE;
  g_cond_signal (gl1_window->context_creation_cond);
  g_mutex_unlock (gl1_window->context_creation_mutex);

  /* Wait the application to call start_loop */
  g_mutex_lock (gl1_window->start_loop_mutex);
  if (!gl1_window->loop_started)
    g_cond_wait (gl1_window->start_loop_cond, gl1_window->start_loop_mutex);
  g_mutex_unlock (gl1_window->start_loop_mutex);

  /* Context and loop creation */
  gl1_window->context = g_main_context_new ();
  gl1_window->loop = g_main_loop_new (gl1_window->context, FALSE);
  
  /* Attach the event handling source. It uses a GPoll depending on the
   * backend. */
  source = g_source_new (&pgm_render_gl1_event_funcs,
                         sizeof (PgmRenderEventSource));
  if (gl1_window->backend->event_poll_fd)
    {
      event_source = (PgmRenderEventSource *) source;
      event_source->event_poll_fd.fd = gl1_window->backend->event_poll_fd;
      event_source->event_poll_fd.events = G_IO_IN;
      g_source_add_poll (source, &event_source->event_poll_fd);
    }
  g_source_set_callback
    (source, (GSourceFunc) gl1_window->backend->dispatch_events,
     gl1_window->backend, NULL);
  g_source_attach (source, gl1_window->context);
  g_source_unref (source);

  /* Attach the rendering callback source */
  gl1_window->source = g_timeout_source_new (4);
  /* gl1_window->source = g_idle_source_new (); */
  g_source_set_callback (gl1_window->source, render, gl1_window, NULL);
  gl1_window->source_id = g_source_attach (gl1_window->source,
                                           gl1_window->context);
  /* g_source_unref (source); */

  /* Start the event handling and rendering loop */
  g_main_loop_run (gl1_window->loop);

  return 0;
}

PgmRenderGl1Window *
pgm_render_gl1_window_new (void)
{
  PgmRenderBaseWindow *window;
  PgmRenderGl1Window *gl1_window;
  GError *error = NULL;

  gl1_window = PGM_RENDER_GL1_WINDOW
    (g_object_new (PGM_RENDER_TYPE_GL1_WINDOW, NULL));
  window = PGM_RENDER_BASE_WINDOW (gl1_window);

  /* FIXME: Using GObject properties we could get a size given by the
   *        user in the constructor. */
  window->width = 800;
  window->height = 600;
  window->canvas_width = 800;
  window->canvas_height = 600;

  gl1_window->loop_thread = g_thread_create (render_loop, gl1_window, TRUE,
					     &error);
  if (error != NULL)
    {
      g_warning ("Could not create thread. Reason: %s", error->message);
      return NULL;
    }

  g_mutex_lock (gl1_window->context_creation_mutex);
  if (!gl1_window->context_created)
    g_cond_wait (gl1_window->context_creation_cond,
                 gl1_window->context_creation_mutex);
  g_mutex_unlock (gl1_window->context_creation_mutex);

  return gl1_window;
}

static gboolean
pgm_render_gl1_event_prepare (GSource *source, gint *timeout)
{
  *timeout = -1;

  if (g_window->backend->event_poll_fd)
    return FALSE;
  else
    return g_window->backend->check_events (g_window->backend);
}

static gboolean
pgm_render_gl1_event_check (GSource *source)
{
  if (g_window->backend->event_poll_fd)
    {
      PgmRenderEventSource *event_source = (PgmRenderEventSource *) source;
      return ((event_source->event_poll_fd.revents & G_IO_IN) != 0);
    }
  else
    return g_window->backend->check_events (g_window->backend);
}

static gboolean
pgm_render_gl1_event_dispatch (GSource *source, GSourceFunc callback,
			       gpointer data)
{
  if (callback)
    return callback (data);

  return FALSE;
}

static PgmRenderError
pgm_render_gl1_window_get_events (PgmRenderBaseWindow *window, GList **event)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  *event = g_list_reverse (gl1_window->event);
  gl1_window->event = NULL;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_start_loop (PgmRenderBaseWindow *window)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  g_mutex_lock (gl1_window->start_loop_mutex);
  gl1_window->loop_started = TRUE;
  g_cond_signal (gl1_window->start_loop_cond);
  g_mutex_unlock (gl1_window->start_loop_mutex);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_quit_loop (PgmRenderBaseWindow *window)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  PGM_RENDER_BASE_WINDOW_LOCK (window);
  g_source_remove (gl1_window->source_id);
  g_source_destroy (gl1_window->source);
  g_main_loop_quit (gl1_window->loop);
  gl1_window->loop_started = FALSE;
  PGM_RENDER_BASE_WINDOW_UNLOCK (window);
  render (window);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_update (PgmRenderBaseWindow *window)
{
  PGM_RENDER_BASE_WINDOW_LOCK (window);
  window->dirty = TRUE;
  PGM_RENDER_BASE_WINDOW_UNLOCK (window);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_set_color (PgmRenderBaseWindow *window, guchar r,
				 guchar g, guchar b, guchar a)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  gl1_window->color.r = (gfloat) r * INV_255;
  gl1_window->color.g = (gfloat) g * INV_255;
  gl1_window->color.b = (gfloat) b * INV_255;
  gl1_window->color.a = (gfloat) a * INV_255;

  gl1_window->new_refresh_color = TRUE;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_get_color (PgmRenderBaseWindow *window, guchar *r,
				 guchar *g, guchar *b, guchar *a)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  g_return_val_if_fail (r, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (g, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (b, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (a, PGM_RENDER_ERROR_X);

  *r = (guchar) (gl1_window->color.r * 255);
  *g = (guchar) (gl1_window->color.g * 255);
  *b = (guchar) (gl1_window->color.b * 255);
  *a = (guchar) (gl1_window->color.a * 255);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_set_title (PgmRenderBaseWindow *window,
				 const gchar *window_title,
				 const gchar *icon_title)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  g_free (gl1_window->window_title);
  gl1_window->window_title = g_strndup (window_title, strlen (window_title));

  g_free (gl1_window->icon_title);
  gl1_window->icon_title = g_strndup (icon_title, strlen (icon_title));

  gl1_window->new_title = TRUE;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_set_cursor (PgmRenderBaseWindow *window, 
                                  PgmRenderCursor cursor)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  gl1_window->new_cursor = TRUE;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_set_size (PgmRenderBaseWindow *window, gint width,
				gint height)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  gl1_window->new_size = TRUE;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_set_canvas_size (PgmRenderBaseWindow *window,
                                       gint canvas_width, gint canvas_height)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  if (window->scale_mode != PGM_RENDER_WINDOW_SCREEN_ADJUST)
    gl1_window->new_canvas_size = TRUE;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_get_screen_resolution (PgmRenderBaseWindow *window,
                                             gint *width, gint *height)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height, PGM_RENDER_ERROR_X);

  *width = gl1_window->backend->resolution_width;
  *height = gl1_window->backend->resolution_height;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_get_screen_size (PgmRenderBaseWindow *window,
                                       gint *width, gint *height)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height, PGM_RENDER_ERROR_X);

  /* FIXME: Return the right size */
  *width = gl1_window->backend->resolution_width;
  *height = gl1_window->backend->resolution_height;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_get_screen_size_mm (PgmRenderBaseWindow *window,
                                          gint *width, gint *height)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height, PGM_RENDER_ERROR_X);

  *width = gl1_window->backend->screen_width_mm;
  *height = gl1_window->backend->screen_height_mm;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_set_fullscreen (PgmRenderBaseWindow *window,
				      gboolean fullscreen, gint width,
                                      gint height)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  window->width = width;
  window->height = height;

  if (gl1_window->fullscreen != fullscreen)
    {
      gl1_window->fullscreen = fullscreen;
      gl1_window->fullscreen_request = TRUE;
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_get_fullscreen (PgmRenderBaseWindow *window,
				      gboolean *fullscreen)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  g_return_val_if_fail (fullscreen, PGM_RENDER_ERROR_X);

  *fullscreen = gl1_window->fullscreen;

  return PGM_RENDER_ERROR_X;
}

static void
_set_scale (PgmRenderGl1Window *gl1_window, gfloat sx, gfloat sy)
{
  PgmRenderBaseWindow *window = PGM_RENDER_BASE_WINDOW (gl1_window);
  GList *element;

  window->sx = sx;
  window->sy = sy;

  /* Scale the surfaces */
  element = gl1_window->surface;
  while (element)
    {
      pgm_render_gl1_surface_set_scale
        (PGM_RENDER_GL1_SURFACE (element->data), sx, sy);
      element = g_list_next (element);
    }
  /* Scale the fonts */
  element = gl1_window->font;
  while (element)
    {
      pgm_render_gl1_font_set_scale
        (PGM_RENDER_GL1_FONT (element->data), sx, sy);
      element = g_list_next (element);
    }
}

static PgmRenderError
pgm_render_gl1_window_set_scale (PgmRenderBaseWindow *window, gfloat sx,
                                 gfloat sy)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);

  if (window->scale_mode == PGM_RENDER_WINDOW_CUSTOM)
    _set_scale (gl1_window, sx, sy);

  return PGM_RENDER_ERROR_X;
}

/* static PgmRenderError */
/* pgm_render_gl1_window_set_aspect_correction_factor (PgmRenderBaseWindow *window, */
/*                                                     gfloat width_factor, */
/*                                                     gfloat height_factor) */
/* {   */
/*   return PGM_RENDER_ERROR_X; */
/* } */

static PgmRenderError
pgm_render_gl1_window_set_scale_mode (PgmRenderBaseWindow *window,
                                      PgmRenderWindowScaleMode scale_mode)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);
  gfloat resolution_width, resolution_height, screen_width_mm, screen_height_mm;
  const gchar *env_var;
  gfloat factor;

  if (scale_mode == PGM_RENDER_WINDOW_SCREEN_ADJUST)
    {
      /* FIXME: Temporary hack to retrieve the physical screen size
       *        before the window creation. The good way to solve that
       *        is to add construction/destruction callbacks in the
       *        plugin system. */
#ifdef G_OS_UNIX
      Display *dpy;
      gint screen;

      dpy = XOpenDisplay (NULL);
      if (!dpy)
        {
          g_print ("Unable to open default display");
          return PGM_RENDER_ERROR_X;
        }

      screen = DefaultScreen (dpy);
      resolution_width = DisplayWidth (dpy, screen);
      resolution_height = DisplayHeight (dpy, screen);
      screen_width_mm = DisplayWidthMM (dpy, screen);
      screen_height_mm = DisplayHeightMM (dpy, screen);

      factor= (screen_width_mm / screen_height_mm)
        * (resolution_height / resolution_width);
#else
      factor = 1.0f;
#endif /* G_OS_UNIX */

      env_var = g_getenv (BROKEN_X_SIZE_ENV_NAME);
      if (env_var && env_var[0] == '1')
        factor = 1.33f;

      window->canvas_width *= factor;
      _set_scale (gl1_window,
                  (gfloat) window->width / (gfloat) window->canvas_width,
                  (gfloat) window->height / (gfloat) window->canvas_height);
    }

  return PGM_RENDER_ERROR_X;
}

static gint
pgm_render_gl1_window_compare (gconstpointer a, gconstpointer b)
{
  gfloat depth_a, depth_b, diff;

  if (PGM_RENDER_IS_GL1_SURFACE (a))
    depth_a = PGM_RENDER_GL1_SURFACE (a)->translation[2];
  else
    depth_a = 0.0f;

  if (PGM_RENDER_IS_GL1_SURFACE (b))
    depth_b = PGM_RENDER_GL1_SURFACE (b)->translation[2];
  else
    depth_b = 0.0f;

  diff = depth_a - depth_b;
  if (diff < 0.0f)
    return -1;
  else
    return 1;
}

static PgmRenderError
pgm_render_gl1_window_add_drawable (PgmRenderBaseWindow *window,
				    PgmRenderDrawable *drawable,
				    PgmRenderDrawableOrder order)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);
  PgmRenderGl1WindowClass *gl1_window_class
    = PGM_RENDER_GL1_WINDOW_GET_CLASS (window);
  gboolean is_surface = PGM_RENDER_IS_GL1_SURFACE (drawable);
  gboolean is_font = PGM_RENDER_IS_GL1_FONT (drawable);

  /* Catch a pointer to the surface class */
  if (!gl1_window_class->surface_class)
    if (is_surface)
      gl1_window_class->surface_class = PGM_RENDER_GL1_SURFACE_GET_CLASS
        (PGM_RENDER_GL1_SURFACE (drawable));

  /* Catch a pointer to the font class */
  if (!gl1_window_class->font_class)
    if (is_font)
      gl1_window_class->font_class = PGM_RENDER_GL1_FONT_GET_CLASS
        (PGM_RENDER_GL1_FONT (drawable));

  /* Add to the surfaces applying the actual window scale factor */
  if (is_surface)
    {
      gl1_window->surface = g_list_append (gl1_window->surface, drawable);
      pgm_render_gl1_surface_set_scale
        (PGM_RENDER_GL1_SURFACE (drawable), window->sx, window->sy);
    }
  /* Add to the fonts applying the actual window scale factor */
  else if (is_font)
    {
      gl1_window->font = g_list_append (gl1_window->font, drawable);
      pgm_render_gl1_font_set_scale
        (PGM_RENDER_GL1_FONT (drawable), window->sx, window->sy);
    }

  /* Add to the far drawables */
  if (order == PGM_RENDER_DRAWABLE_FAR)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->far_drawable = g_list_append (window->far_drawable, drawable);
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
      drawable->order = order;
      drawable->window = window;
    }
  /* Add to the z-ordered drawables */
  else if (order == PGM_RENDER_DRAWABLE_Z_ORDERED)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->z_ordered_drawable = g_list_insert_sorted
	(window->z_ordered_drawable, drawable, pgm_render_gl1_window_compare);
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
      drawable->order = order;
      drawable->window = window;
    }
  /* Add to the near drawables */
  else if (order == PGM_RENDER_DRAWABLE_NEAR)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->near_drawable = g_list_append (window->near_drawable, drawable);
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
      drawable->order = order;
      drawable->window = window;
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_del_drawable (PgmRenderBaseWindow *window,
				    PgmRenderDrawable *drawable)
{
  /* Remove from the far drawables */
  if (drawable->order == PGM_RENDER_DRAWABLE_FAR)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->far_drawable = g_list_remove (window->far_drawable, drawable);
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
      drawable->order = PGM_RENDER_DRAWABLE_UNBINDED;
      drawable->window = NULL;
    }
  /* Remove from the z-ordered drawables */
  else if (drawable->order == PGM_RENDER_DRAWABLE_Z_ORDERED)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->z_ordered_drawable = g_list_remove (window->z_ordered_drawable,
						  drawable);
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
      drawable->order = PGM_RENDER_DRAWABLE_UNBINDED;
      drawable->window = NULL;
    }
  /* Remove from the near drawables */
  else if (drawable->order == PGM_RENDER_DRAWABLE_NEAR)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->near_drawable = g_list_remove (window->near_drawable, drawable);
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
      drawable->order = PGM_RENDER_DRAWABLE_UNBINDED;
      drawable->window = NULL;
    }

  return PGM_RENDER_ERROR_X;
}

static gint
reorder_drawable (PgmRenderBaseWindow *window, GList **list,
		  PgmRenderDrawable *drawable, gint position)
{
  GList *old_link;
  GList *new_link;
  PgmRenderDrawable *info = NULL;
  gint old_position;

  old_link = *list;
  old_position = 0;
  while (old_link)
    {
      info = old_link->data;
      if (info == drawable)
	break;

      old_link = old_link->next;
      old_position++;
    }

  g_return_val_if_fail (old_link != NULL, 2);

  if (position == old_position)
    return 1;

  PGM_RENDER_BASE_WINDOW_LOCK (window);
  *list = g_list_delete_link (*list, old_link);
  if (position < 0)
    new_link = NULL;
  else
    new_link = g_list_nth (*list, position);
  *list = g_list_insert_before (*list, new_link, info);
  PGM_RENDER_BASE_WINDOW_UNLOCK (window);

  return 0;
}

static PgmRenderError
pgm_render_gl1_window_reorder_drawable (PgmRenderBaseWindow *window,
					PgmRenderDrawable *drawable,
					gint position)
{
  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  /* The drawable is in the far drawables */
  if (drawable->order == PGM_RENDER_DRAWABLE_FAR)
    reorder_drawable (window, &window->far_drawable, drawable, position);
  /* The drawable is in the z-ordered drawables */
  else if (drawable->order == PGM_RENDER_DRAWABLE_Z_ORDERED)
    {
      PGM_RENDER_BASE_WINDOW_LOCK (window);
      window->z_ordered_drawable = g_list_remove (window->z_ordered_drawable,
						  drawable);
      window->z_ordered_drawable = g_list_insert_sorted
	(window->z_ordered_drawable, drawable, pgm_render_gl1_window_compare);
      PGM_RENDER_BASE_WINDOW_UNLOCK (window);
    }
  /* The drawable is in the near drawables */
  else if (drawable->order == PGM_RENDER_DRAWABLE_NEAR)
    reorder_drawable (window, &window->near_drawable, drawable, position);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
pgm_render_gl1_window_regenerate_drawables (PgmRenderBaseWindow *window)
{
  PgmRenderGl1Window *gl1_window = PGM_RENDER_GL1_WINDOW (window);
  GList *font;

  /* ATM, just the fonts need to be regenerated */
  font = gl1_window->font;
  while (font)
    {
      pgm_render_gl1_font_regenerate (PGM_RENDER_GL1_FONT (font->data));
      font = g_list_next (font);
    }

  return PGM_RENDER_ERROR_X;
}
